#include <linux/types.h>
#include <linux/errno.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/fcntl.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/init.h>
#include <linux/mtd/mtd.h>

#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/atomic.h>
#include <asm/avalanche/generic/adam2_env.h>
#include <asm/avalanche/generic/ticfg.h>
#include <asm/avalanche/generic/nsp_single_image.h>

/*
#define _TICFG_DEBUG_
*/

#ifdef _TICFG_DEBUG_
#define DPRINTK(args...) do { printk(args); } while(0);
#else
#define DPRINTK(args...) do { } while(0);
#endif

#define TICFG_DEVNAME        "ticfg"
#define TICFG_ENV_DIR        "env"

static void ticfg_mtd_notify_add(struct mtd_info* mtd);
static void ticfg_mtd_notify_remove(struct mtd_info* mtd);
static struct mtd_notifier notifier = {
	add:	ticfg_mtd_notify_add,
	remove:	ticfg_mtd_notify_remove,
};

static int ticfg_open(struct inode *inode, struct file *filp);
static int ticfg_release(struct inode *inode, struct file *filp);
static ssize_t ticfg_read(struct file *filp, char *buff,
			  size_t count, loff_t *offp);
static ssize_t ticfg_write(struct file *filp, const char *buff,
			   size_t count, loff_t *offp);
static int ticfg_ioctl(struct inode *inode, struct file *file,
		     u_int cmd, u_long arg);

typedef struct {
	struct mtd_info *config;	/* Config space */
	struct mtd_info *secondary_config;	/* Secondary Config space */
	struct mtd_info *primary_image;
	struct mtd_info *secondary_image;
	int count;
	int size;
	wait_queue_head_t wq;
} ticfg_dev_t;

static ticfg_dev_t ticfg_dev;

static devfs_handle_t devfs_handle = NULL;

static atomic_t ticfg_busy = ATOMIC_INIT(0);
static spinlock_t ticfg_busy_lock = SPIN_LOCK_UNLOCKED;

static struct proc_dir_entry *ticfg_proc_dir;
static struct proc_dir_entry *ticfg_proc_env_dir;

struct file_operations ticfg_fops = {
	open: ticfg_open,
	read: ticfg_read,
	write: ticfg_write,
	release: ticfg_release,
	ioctl: ticfg_ioctl
};

/* Acquire the mutual exclusion lock for the configuration region. 
 *
 * We do it this way because we do not know which context it will be
 * called from and because we need to allow sleepable calls when we
 * own the lock.  These requirements rule out the usual spinlocks and
 * semaphores.
 */
void ticfg_lock_region(void)
{
	unsigned long flags;
	int available=0;

	while(!available) {
		spin_lock_irqsave(&ticfg_busy_lock, flags);
		if(!atomic_read(&ticfg_busy)) {
			available=1;
			atomic_set(&ticfg_busy,1);
		}
		spin_unlock_irqrestore(&ticfg_busy_lock, flags);
	}
}

/* Release the mutual exclusion lock for the configuration region */
void ticfg_unlock_region(void)
{
	unsigned long flags;

	spin_lock_irqsave(&ticfg_busy_lock, flags);
	atomic_set(&ticfg_busy,0);
	spin_unlock_irqrestore(&ticfg_busy_lock, flags);

}

static void mtd_erase_callback (struct erase_info *instr)
{
	DPRINTK("ticfg: mtd_erase_callback\n");
	wake_up((wait_queue_head_t *)instr->priv);
}

static int ticfg_erase_mtd_region(struct mtd_info *mtd, unsigned long offset, unsigned long len)
{
	struct erase_info *erase;

	int ret=0;
	DECLARE_WAITQUEUE(wait,current);
	wait_queue_head_t wait_q;

	DPRINTK("ticfg: mtd_erase_mtd_region(offset=%lu len=%lu)\n", offset, len);

	init_waitqueue_head(&wait_q);
	
	erase=kmalloc(sizeof(struct erase_info), GFP_KERNEL);
	if (!erase) 
		return -ENOMEM;
	memset(erase,0,sizeof(struct erase_info));
	
	erase->mtd=mtd;
	erase->callback=mtd_erase_callback;
	erase->addr=offset;
	erase->len=len;
	erase->priv=(u_long)&wait_q;

	ret=mtd->erase(mtd, erase);
	if( !ret ) {
		set_current_state(TASK_UNINTERRUPTIBLE);
		add_wait_queue( &wait_q, &wait);
		if( erase->state!=MTD_ERASE_DONE &&
		    erase->state!=MTD_ERASE_FAILED) {
			schedule();
		}
		
		remove_wait_queue(&wait_q, &wait);
		set_current_state(TASK_RUNNING);
		
		ret=(erase->state==MTD_ERASE_FAILED)?-EIO:0;
	} else {
		printk(KERN_ERR "Failed to erase mtd, region [0x%x,0x%x]\n", 
		       erase->addr,erase->len);
	}
	
	kfree(erase);

	return ret;
}


static int ticfg_read_mtd_region(struct mtd_info *mtd, off_t offset, size_t count, 
				 size_t *bytes_read, char *kbuf)
{
	int ret=0;
	int retlen;

	DPRINTK("ticfg: read_mtd_region() offset=%lu count=%lu\n", (u_long) offset, (u_long) count);

	*bytes_read=0;
	while(count) {
		ret=MTD_READ(mtd, offset, count, &retlen, kbuf);
		if(!ret) {
			offset+=retlen;
			count-=retlen;
			kbuf+=retlen;
			*bytes_read+=retlen;
		}
		else {
			return ret;
		}
	}
	
	return 0;		
	
}


static int ticfg_write_mtd_region(struct mtd_info *mtd, off_t offset, size_t count, 
			    size_t *bytes_written, char *kbuf)
{
	int ret=0;
	int retlen;
	
	DPRINTK("ticfg: write_mtd_region() offset=%lu count=%lu\n", (u_long) offset, (u_long) count);
                
	*bytes_written=0;
	while (count) {
		ret = MTD_WRITE(mtd, offset, count, &retlen, kbuf);
                if (!ret) {
                        offset += retlen;
                        count -= retlen;
                        kbuf += retlen;
                        *bytes_written += retlen;
		} else {
			return ret;
                }
        }

	return 0;
}

/* Read from the cfgman area of the configuration region */
static ssize_t ticfg_read(struct file *filp, char *buf, size_t count, loff_t *ppos)
{
	ticfg_dev_t *dev=(ticfg_dev_t *)filp->private_data;
	size_t retlen=0;
	char *kbuf;
	int ret=0;
	off_t offset;

	DPRINTK("ticfg: read() offset=%lu count=%lu\n", (u_long) *ppos, (u_long)  count);

	if( *ppos+count > CFGMAN_MTD_SIZE )
		count=CFGMAN_MTD_SIZE-*ppos;

	if( !count) return 0;

	kbuf=kmalloc(count,GFP_KERNEL);
	if(!kbuf) {
		return -ENOMEM;
	}

	ticfg_lock_region();

	offset=*ppos+CFGMAN_MTD_OFFSET;
	ret=ticfg_read_mtd_region(dev->config, offset, count, &retlen, kbuf);
	if(ret) {	
		ticfg_unlock_region();
		kfree(kbuf);
		return ret;
	}

	ticfg_unlock_region();

	*ppos+=retlen;
	if( copy_to_user(buf, kbuf, retlen)) {
		kfree(kbuf);
		return -EFAULT;
	}
 
	kfree(kbuf);
	return retlen;
}



/* Write to the cfgman area of the configuration region. */
static ssize_t ticfg_write(struct file *filp, const char *buf,
			   size_t count, loff_t *ppos)
{
	ticfg_dev_t *dev=(ticfg_dev_t *)filp->private_data;
	size_t retlen=0;
	int ret=0;
	off_t offset;

	DPRINTK("ticfg: write(0x%08lx, 0x%08lx, %lu, %lu)\n",
		(u_long) filp, (u_long) buf, (u_long) count,(u_long) *ppos);

	if( *ppos+count > CFGMAN_MTD_SIZE )
		count=CFGMAN_MTD_SIZE-*ppos;

	if( !count) return 0;

	ticfg_lock_region();

	offset=*ppos+CFGMAN_MTD_OFFSET;
	ret=ticfg_write_mtd_region(dev->config, offset, count, &retlen, (char *) buf);
	if(ret) {
		ticfg_unlock_region();
		return ret;
	}
	
	ticfg_unlock_region();
	
	*ppos+=retlen;
	return retlen;
}

/* Erase the cfgman area in the configuration region */
int ticfg_erase_cfgman(struct mtd_info *mtd)
{
	char *kbuf;
	int ret;
	int retlen;

	DPRINTK("ticfg: erase_cfgman()\n");

 	kbuf=kmalloc(ENV_VAR_MTD_SIZE,GFP_KERNEL);
	if(!kbuf)
		return -ENOMEM;

	ticfg_lock_region();

	/* save the current set of environment variables */
	ret=ticfg_read_mtd_region(mtd, ENV_VAR_MTD_OFFSET, ENV_VAR_MTD_SIZE, &retlen, kbuf);
	if(ret) {
		ticfg_unlock_region();
		kfree(kbuf);
		return ret;
	}

	/* erase the entire region */
	ret=ticfg_erase_mtd_region(mtd, TICFG_MTD_OFFSET, TICFG_MTD_SIZE);
	if(ret) {
		ticfg_unlock_region();
		kfree(kbuf);
		return ret;
	}

	/* restore the current set of environment variables */
	ret=ticfg_write_mtd_region(mtd, ENV_VAR_MTD_OFFSET, ENV_VAR_MTD_SIZE, &retlen, kbuf);
	if(ret) {
		ticfg_unlock_region();
		kfree(kbuf);
		return ret;
	}	

	ticfg_unlock_region();
	kfree(kbuf);
	return 0;
}

/* Erase the environment variables area in the configuration region */
int ticfg_erase_env_vars(struct mtd_info *mtd)
{
	char *kbuf;
	int ret;
	int retlen;

	DPRINTK("ticfg: erase_envars()\n");

 	kbuf=kmalloc(CFGMAN_MTD_SIZE,GFP_KERNEL);
	if(!kbuf)
		return -ENOMEM;

	ticfg_lock_region();

	/* save the current configman configuration */
	ret=ticfg_read_mtd_region(mtd, CFGMAN_MTD_OFFSET, CFGMAN_MTD_SIZE, &retlen, kbuf);
	if(ret) {
		ticfg_unlock_region();
		kfree(kbuf);
		return ret;
	}	

	/* erase the entire region */
	ret=ticfg_erase_mtd_region(mtd, TICFG_MTD_OFFSET, TICFG_MTD_SIZE);
	if(ret) {
		ticfg_unlock_region();
		kfree(kbuf);
		return ret;
	}

	/* restore the current configman configuration */
	ret=ticfg_write_mtd_region(mtd, CFGMAN_MTD_OFFSET, CFGMAN_MTD_SIZE, &retlen, kbuf);
	if(ret) {
		ticfg_unlock_region();
		kfree(kbuf);
		return ret;
	}	
	
	ticfg_unlock_region();
	kfree(kbuf);
	return ret;
}

/* Set the boot flags. Remember, this is flash. We set to 0. We cannot
   clear without formating that flash sector */
static int ticfg_ioctl_flags(int secondary, int write, unsigned int mask, unsigned int *flags)
{
	struct nsp_img_hdr	hdr;	/* We don't worry about the whole header, just
					   the first part */
	struct mtd_info *mtd;
	unsigned int inv_flags;
	int	retlen;
	int	ret;

	if(!flags) return -1;

	/* Get the mtd info block */
	mtd=ticfg_dev.primary_image;
	if(secondary)
		mtd=ticfg_dev.secondary_image;

	/* Get access to the header, and do some validation checks */
	ret=MTD_READ(mtd, 0, sizeof(hdr), &retlen, (char*)&hdr);
	if(retlen != sizeof(hdr) || hdr.head.magic != NSP_IMG_MAGIC_NUMBER)
		return -1;	/* Not a single image */

	/* Only allow us to modify the failback flags */
	mask &= NSP_IMG_FLAG_FAILBACK_MASK;
	*flags = ~(hdr.head.flags) & mask;	/* We SET to 0, so we invert the whole
						   flags field */
	if(!write)	return 0; /* If read, just return now */

	inv_flags = ~(mask);
	ret = MTD_WRITE(mtd, 2*sizeof(unsigned int), sizeof(unsigned int), &retlen, (char*)&inv_flags);
	if(retlen != sizeof(unsigned int))
		return -1;

	/* Reread the flags field */
	ret=MTD_READ(mtd, 0, sizeof(hdr), &retlen, (char*)&hdr);
	if(retlen != sizeof(hdr) || hdr.head.magic != NSP_IMG_MAGIC_NUMBER)
		return -1;	/* Not a single image */
	*flags = ~(hdr.head.flags);

	return 0;
}

static int ticfg_ioctl(struct inode *inode, struct file *filp,
		     u_int cmd, u_long arg)
{
	ticfg_dev_t *dev=(ticfg_dev_t *)filp->private_data;
	int ret=0;

	DPRINTK("ticfg: ioctl(cmd 0x%08x, arg 0x%08lx)\n", cmd, arg);

	switch(cmd) {
	case TICFG_IOCTL_ERASE:
	{
		ret=ticfg_erase_cfgman(dev->config);
		if(ret)
			return ret;

		break;
	}
	case TICFG_IOCTL_FLAG_READ_PRIMARY:
	{
		unsigned int	mask;
		unsigned int	flags;
		ret=__get_user(mask, (unsigned int *)arg);
		ticfg_ioctl_flags(0, 0, mask, &flags);
		ret=__put_user(flags,(unsigned int *)arg);
		break;
	}
	case TICFG_IOCTL_FLAG_SET_PRIMARY:
	{
		unsigned int	mask;
		unsigned int	flags;
		ret=__get_user(mask, (unsigned int *)arg);
		ticfg_ioctl_flags(0, 1, mask, &flags);
		ret=__put_user(flags,(unsigned int *)arg);
		break;
	}
	case TICFG_IOCTL_FLAG_READ_SECONDARY:
	{
		unsigned int	mask;
		unsigned int	flags;
		ret=__get_user(mask, (unsigned int *)arg);
		ticfg_ioctl_flags(1, 0, mask, &flags);
		ret=__put_user(flags,(unsigned int *)arg);
		break;
	}
	case TICFG_IOCTL_FLAG_SET_SECONDARY:
	{
		unsigned int	mask;
		unsigned int	flags;
		ret=__get_user(mask, (unsigned int *)arg);
		ticfg_ioctl_flags(1, 1, mask, &flags);
		ret=__put_user(flags,(unsigned int *)arg);
		break;
	}
		break;
	default:
		break;
	}

	return 0;
}

static int ticfg_proc_read_env(char *page, char **start, off_t pos, int count,
			       int *eof, void *data)
{
    int len = 0, i, limit = count-80;
    char *pvar, *pval;
    int break_flag = 0;
    static int last_index = 0;

    if( !pos )
     last_index = 0;

	/* Scan thru the flash, looking for each possible variable index */
    for ( i = last_index; ; i++, last_index++ )
    {
        int tmplen;
        
        /* Get the next variable */
        if( (pvar=adam2_env_get_variable(i))!=NULL ) 
        {
            /* Check if variable has a value */
            if( (pval=adam2_env_get_value(pvar)) != NULL) 
            {
                /* Calculate next line length, store in tmplen */
                tmplen = strlen( pval ) + strlen( pvar ) + 2;
                
                /* Check if there is enough space in the buffer */
                if ( len + tmplen > limit )
                {
                    /* We exceeded data length. */
                    break_flag = 1;
                    
                    /* Free all data */
                    kfree(pval);
                    kfree(pvar);

                    /* Stop the iteration */
                    break;
                }
                else
                {
                    /* Format the next line */
                    len += sprintf(page + len, "%s\t%s\n", pvar, pval);;
                    kfree(pval);
                }
            }

            /* Free variable buffer */
            kfree(pvar);

        }
        else
           break; /* No more variables */
    }
   
    /* Mark EOF if we broke the iteration, or buffer length is smaller than
       the allocated buffer */
    if ( !break_flag )
       *eof = 1;
  
    /* Always */
    *start = page;
       
    if( len > limit )
       len = limit;

	return len;
}

#define MAX_ENV_DATA_LEN (FLASH_ENV_ENTRY_SIZE)
static int ticfg_proc_write_env(struct file *file, const char *buffer, 
				unsigned long count, void *data)
{
	char envdata[MAX_ENV_DATA_LEN];
	char *pvar;
	char *pval;

	/* TODO: Check for count > len */
	if(count > MAX_ENV_DATA_LEN){
		printk(KERN_WARNING "Truncating environment variable value.\n");
		count=MAX_ENV_DATA_LEN;
	}
	if( copy_from_user(envdata, buffer, count) ) {
		return -EFAULT;
	}

	/* extract the variable/value pair */
	if( count )
		envdata[count-1]='\0';
	else
		envdata[count]='\0';
	
	pvar=envdata;
	pval=strpbrk(envdata," \t");
	if(pval) *pval++='\0';

	if(pval) {
		/* write the variable/value pair to flash */
		if(adam2_env_set_variable(pvar, pval) != 0)
		{
			printk(KERN_WARNING "Defragginig environment variable region.\n");
			adam2_env_defrag();
			if( adam2_env_set_variable(pvar, pval) != 0 )
				printk(KERN_ERR "Failed to write %s to environment variable region.\n", pvar);
		}

	} else {
		/* We have no way to distinguish between unsetting a
		 * variable and setting a non-value variable.
		 *
		 * Consequently, we will treat any variable
		 * without a value as an unset request.
		 */
		adam2_env_unset_variable(pvar);
	}	

	return count;
}


static int ticfg_open(struct inode *inode, struct file *filp) 
{
	ticfg_dev_t *dev;

	DPRINTK("ticfg: open\n");

	MOD_INC_USE_COUNT;
	
	dev=(ticfg_dev_t *)filp->private_data;
	if(!dev) {
		dev=&ticfg_dev;
		filp->private_data=dev;
	}

	/* Fallback to the old method if we don't have single image */
	if(!dev->config){
		dev->config=get_mtd_device(NULL,CFGMAN_MTD);
		if(!dev->config) {
                printk(KERN_ERR "ticfg: can't get mtd%d\n", CFGMAN_MTD);
		return -EIO;
        }
	}
	return 0;
}

static int ticfg_release(struct inode *inode, struct file *filp)
{
	ticfg_dev_t *dev=(ticfg_dev_t *)filp->private_data;

	DPRINTK("ticfg: release\n");

	if( dev->config->sync ) 
		dev->config->sync(dev->config);

	put_mtd_device(dev->config);
	
	return 0;
}

static void ticfg_mtd_notify_add(struct mtd_info* mtd)
{
	if(!mtd)
		return;
	if(!strcmp("RESERVED_PRIMARY_IMAGE", mtd->name)) {
		ticfg_dev.primary_image=mtd;
	} else if(!strcmp("RESERVED_SECONDARY_IMAGE", mtd->name)) {
		ticfg_dev.secondary_image=mtd;
	} else if(!strcmp("RESERVED_PRIMARY_XML_CONFIG", mtd->name)) {
		ticfg_dev.config=mtd;
	} else if(!strcmp("RESERVED_SECONDARY_XML_CONFIG", mtd->name)) {
		ticfg_dev.secondary_config=mtd;
	}
}

static void ticfg_mtd_notify_remove(struct mtd_info* mtd)
{
}

static int __init ticfg_init(void)
{
	DPRINTK("ticfg: init\n");

	memset( &ticfg_dev, 0, sizeof(ticfg_dev_t));
	devfs_handle=devfs_register(NULL, TICFG_DEVNAME, DEVFS_FL_AUTO_DEVNUM,
				    0,0,
				    S_IFCHR | S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP, 
				    &ticfg_fops, &ticfg_dev);
	
	ticfg_proc_dir=proc_mkdir(TICFG_DEVNAME,NULL);
	if( ticfg_proc_dir==NULL ) {
		printk(KERN_ERR "ticfg: Unable to create /proc/%s entry\n", TICFG_DEVNAME);
		return -ENOMEM;
	}
	
	ticfg_proc_env_dir=create_proc_entry(TICFG_ENV_DIR, 0644,ticfg_proc_dir);
	if( ticfg_proc_env_dir==NULL ) {
		printk(KERN_ERR "TICFG: Unable to create /proc/%s/%s entry\n", TICFG_DEVNAME, TICFG_ENV_DIR);
		remove_proc_entry("ticfg",NULL);
		return -ENOMEM;
	}
	ticfg_proc_env_dir->read_proc=ticfg_proc_read_env;
	ticfg_proc_env_dir->write_proc=ticfg_proc_write_env;

	register_mtd_user(&notifier);

	return 0;
}

static void ticfg_cleanup(void)
{
	devfs_unregister(devfs_handle);

	remove_proc_entry(TICFG_ENV_DIR,ticfg_proc_dir);
	remove_proc_entry(TICFG_DEVNAME, NULL);
}

module_init(ticfg_init);
module_exit(ticfg_cleanup);
