/*******************************************************************************
**+--------------------------------------------------------------------------+**
**|                            ****                                          |**
**|                            ****                                          |**
**|                            ******o***                                    |**
**|                      ********_///_****                                   |**
**|                      ***** /_//_/ ****                                   |**
**|                       ** ** (__/ ****                                    |**
**|                           *********                                      |**
**|                            ****                                          |**
**|                            ***                                           |**
**|                                                                          |**
**|         Copyright (c) 1998-2005 Texas Instruments Incorporated           |**
**|                        ALL RIGHTS RESERVED                               |**
**|                                                                          |**
**+--------------------------------------------------------------------------+**
*******************************************************************************/

/*******************************************************************************
 * FILE PURPOSE:    Graphical LCD Driver and Window Manager Stack Source
 *******************************************************************************
 * FILE NAME:       glcd_hal_apis.c
 *
 * DESCRIPTION:     Graphical LCD HAL APIs Source file
 *
 * REVISION HISTORY:
 * 09-04-2004      Nakshatra     Initial creation
 * 16-02-2005      Hai Shalom
 *
 * (C) Copyright 2004, Texas Instruments, Inc
 ******************************************************************************/

#include "glcd_hal_apis.h"
#include "glcd_hal.h"
#include "glcd_hal.c"

#define COLOR 1
#define COLOR_656 1
#define RASTER_MODE 1
#define RESET     0
#define RUN       1

GLCD_HAL_STATUS lcd_hal_init(const LCD_CNTRL_CONFIG_T *cfg)
{
	UINT32 bpp;

	/* GPIO Init */
  lcd_tb_gpio_init();

	/* Reset the LCD module. */
	LCD_reset();

	/* Configure the LCD clock. */
	SYS_CLK_LCD_CLK_CTRL_REG = cfg->lcd_clk;

	/* Configure the LCD clock divisor. */
	LCD_cfg_clock_divisor(cfg->pxl_clk, LCD_RASTER_MODE);

	/* Configure the DMA burst size. */
	LCD_cfg_dma(cfg->dma_burst_sz);

	/* Configure the AC bias properties. */
	LCD_cfg_ac_bias(cfg->ac_bias, cfg->ac_bias_intrpt);

	/* Configure the vertical and horizontal sync properties. */
	LCD_cfg_vertical_sync(cfg->vbp, cfg->vsw, cfg->vfp);
	LCD_cfg_horizontal_sync(cfg->hbp, cfg->hsw, cfg->hfp);

	/* Configure for disply */
  LCD_cfg_display(cfg);

	if (QVGA != cfg->p_disp_panel->panel_type) {
		os_print(KERN_ALERT"\nError: Only QVGA panel is currently supported !");
		return STATUS_FAILURE;
	}
	if (cfg->bpp <= cfg->p_disp_panel->max_bpp && 
							cfg->bpp >= cfg->p_disp_panel->min_bpp) {
		bpp = cfg->bpp;
	} else {
		bpp = cfg->p_disp_panel->max_bpp;
	}
			
	if( bpp == 12 )
     bpp = 16;
  
  LCD_cfg_frame_buffer((UINT32)cfg->p_disp_panel->width,(UINT32)cfg->p_disp_panel->height, bpp, cfg->raster_order);

	/* Configure FDD */
	LCD_RASTER_CTRL_REG = (LCD_RASTER_CTRL_REG & 0xfff00fff) | 
														(cfg->fdd << 12);

	return STATUS_SUCCESS;
}

/*
 * Synopsis: Blit a whole frame buffer, provided the base address of the same.
 * This routine should stop the current DMA activity, process the input buffer
 * as necessary and start the DMA.
 */
GLCD_HAL_STATUS lcd_hal_blit(RASTER_LOAD_MODE_T load_mode, const char *p_buf)
{
	volatile UINT32 reg;
	volatile UINT32 counter = 0;

	if (LOAD_DATA == load_mode) {
		/* Disable the raster engine. */
		LCD_disable_raster(TRUE);
		while ( !(LCD_STAT_REG & LCD_FRAME_DONE) && (counter < 100000)) {
        	counter++;
    	}

		/* Clear the status. ?? here ?? */
		LCD_clear_status_int(LCD_ALL_STATUS_INTS & 
									~LCD_FRAME_DONE & ~LCD_SYNC_LOST);

		/* Update the databuf in the hw. */
		LCD_DMA_FRM_BUF_BASE_ADDR_0_REG = PHYSICAL_ADDRESS((UINT32)p_buf);
		LCD_DMA_FRM_BUF_CEILING_ADDR_0_REG = PHYSICAL_ADDRESS((UINT32)p_buf
														+ g_databuf_sz - 4);

		/* Start the DMA. */
		reg = LCD_RASTER_CTRL_REG;
		reg &= ~(3 << 20);
		reg |= LCD_PALETTE_LOAD_MODE(DATA_ONLY);
		LCD_RASTER_CTRL_REG = reg | LCD_RASTER_ENABLE |
                           LCD_UNDERFLOW_INT_ENA | LCD_SYNC_LOST_INT_ENA |
                           LCD_LOAD_DONE_INT_ENA | LCD_FRAME_DONE_INT_ENA |
                           LCD_AC_BIAS_INT_ENA;

	LCD_clear_status_int(LCD_FRAME_DONE); /* ?? required here ? */
#ifdef FB_DEBUG
    	LCD_show_raster_status(LCD_STAT_REG);
	/* LCD_show_raster_interface(); */
#endif

		return STATUS_SUCCESS;
	} else if (LOAD_PALETTE == load_mode) {
		/* Disable the raster engine. */
		LCD_disable_raster(TRUE);
		while ( !(LCD_STAT_REG & LCD_FRAME_DONE) && (counter < 100000)) {
        	counter++;
    	}

		/* Clear the status. ?? here ?? */
		LCD_clear_status_int(LCD_ALL_STATUS_INTS & 
									~LCD_FRAME_DONE & ~LCD_SYNC_LOST);

		LCD_DMA_FRM_BUF_BASE_ADDR_0_REG = PHYSICAL_ADDRESS(p_buf);
		LCD_DMA_FRM_BUF_CEILING_ADDR_0_REG = 
					PHYSICAL_ADDRESS((UINT32)p_buf + g_palette_sz - 4);

		reg = LCD_RASTER_CTRL_REG;
        reg &= ~(3 << 20);    /* TODO: NEED EQUATE FOR THIS MAGIC NUMBER! */
        reg |= LCD_PALETTE_LOAD_MODE(PALETTE_ONLY);
        LCD_RASTER_CTRL_REG = reg | LCD_RASTER_ENABLE;

	LCD_RASTER_CTRL_REG &= ~LCD_RASTER_ENABLE;
        LCD_wait_status_int(LCD_PALETTE_LOADED);
        LCD_wait_status_int(LCD_SYNC_LOST);
#ifdef FB_DEBUG
    	LCD_show_raster_status(LCD_STAT_REG);
	/* LCD_show_raster_interface(); */
#endif
		return STATUS_SUCCESS;
	} else {
		os_print(KERN_ALERT "\r\nError: unsupported raster load mode !");
		return STATUS_ERROR;
	}
}

GLCD_HAL_STATUS lcd_ioctl(LCD_CMD_T cmd, void *val)
{
    UINT32 reg, temp;
	GLCD_HAL_STATUS status = STATUS_SUCCESS;

	switch (cmd) {
		case LCD_GET_AC_BIAS_FREQ:
    		reg = LCD_RASTER_TIMING_2_REG & 0xFF00;
			*(UINT32 *)val = reg >> 8;
			break;
		case LCD_SET_AC_BIAS_FREQ:
			temp = *(UINT32 *)val;
			if (temp > 0xFF) temp = 0xFF;
    		reg = LCD_RASTER_TIMING_2_REG & 0xFFFF0000;
    		reg |= LCD_AC_BIAS_FREQUENCY(temp);
    		LCD_RASTER_TIMING_2_REG = reg;
			break;
		case LCD_GET_SCREEN_WIDTH:
			reg  = (LCD_RASTER_TIMING_0_REG & 0x3F0) >> 4;
			*(UINT32 *)val = (reg + 1) << 4;
			break;
		case LCD_GET_SCREEN_HEIGHT:
			reg  = LCD_RASTER_TIMING_1_REG & 0x3FF;
			*(UINT32 *)val = reg + 1;
			break;
		default:
			status = STATUS_ERR_UNKNOWN_CMD;
			break;
	}
	return status;
}

/* Enable the LCD Controller Hardware */
void hardware_enable(void)
{
	volatile UINT32 reg;
        
	/* Initialize the GPIO Pins of LCD (Tri-State to Active) */
	lcd_tb_gpio_init();

	/* Enable the Clock of LCD Controller */
	reg = SYS_CLK_DSBL_CTRL_REG;
  reg 	&= (~LCD_CLK_DSBL);
  SYS_CLK_DSBL_CTRL_REG = reg;

#ifdef FB_DEBUG
	os_print(KERN_INFO "GLCD: LCD Hardware Enabled\n");
#endif
}


/* Disable the LCD Controller Hardware */
void hardware_disable(void)
{
	volatile UINT32 reg;

	/* Disable the GPIO Pins (LCD Pins to Tri-State) */
        reg = SYS_RESET_PIN_SEL_REG(8);
        reg     &= (~(0x44000000));
        SYS_RESET_PIN_SEL_REG(8) = reg;

        reg = SYS_RESET_PIN_SEL_REG(9);
        reg     &= (~(0x55145515));
        SYS_RESET_PIN_SEL_REG(9) = reg;

        reg = SYS_RESET_PIN_SEL_REG(10);
        reg     &= (~(0x00005505));
	SYS_RESET_PIN_SEL_REG(10) = reg;

	/* Disable the Clock to LCD Controller */
	reg = SYS_CLK_DSBL_CTRL_REG;
        reg 	|= LCD_CLK_DSBL;
        SYS_CLK_DSBL_CTRL_REG = reg;

#ifdef FB_DEBUG
	os_print(KERN_INFO "GLCD: LCD Hardware Disabled\n");
#endif
}


/* To Check whether LCD Controller is Active or Not */
int lcd_active(void)
{
	/* Return True if LCD Controller is Running */
	volatile UINT32 reg;
	reg = LCD_RASTER_CTRL_REG;
	return (reg & 1);
}


/* Enable the LCD Raster Engine */
void lcd_enable(void)
{
	volatile UINT32 reg;

        /* Enable Raster if previously not Active*/
	if (!lcd_active())
	{
		reg = LCD_RASTER_CTRL_REG;
	        reg &= ~(3 << 20);
		reg |= LCD_PALETTE_LOAD_MODE(PALETTE_AND_DATA);
	        LCD_RASTER_CTRL_REG = reg | LCD_RASTER_ENABLE;
	}

#ifdef FB_DEBUG
	os_print(KERN_INFO "GLCD: LCD Raster Engine Enabled\n");
#endif
}


/* Disable the LCD Raster Engine */
void lcd_disable(void)
{
	UINT32 counter;

	counter = 0;

	/* Disable the Raster and wait till End of Frame detected*/
        LCD_disable_raster(TRUE);
	/* Wait until the active Raster Frame is Complete*/
        while ( !(LCD_STAT_REG & LCD_FRAME_DONE) && (counter < 100000)) {
                counter++;
        }

#ifdef FB_DEBUG
	os_print(KERN_INFO "GLCD: LCD Raster Engine Disabled\n");
#endif
}
