/**************************************************************************
 * FILE PURPOSE	:  	TCA Configuration Control.
 **************************************************************************
 * FILE NAME	:   tcactl.c
 *
 * DESCRIPTION	:
 *  This file implements the configuration control kernel piece that is used
 *  to configure the various IP QoS Framework blocks. 
 *
 *	CALL-INs:
 *
 *	CALL-OUTs:
 *
 *	User-Configurable Items:
 *
 *	(C) Copyright 2005, Texas Instruments, Inc.
 *************************************************************************/

/**************************** INCLUDE FILES *************************/
#include <asm/uaccess.h>
#include <linux/kernel.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/proc_fs.h>
#include <linux/skbuff.h>
#include <linux/ti_qos.h>
#include <linux/module.h>

/**************************** DEFINITIONS *************************/

/* These define the status of the configuration requests. */
#define TCA_SUCCESS                 0
#define TCA_SYNTAX_ERROR            -1
#define TCA_INTERNAL_ERROR          -2

/* This is the max. length of the output section. */
#define MAX_OUTPUT_SECTION_LENGTH   2600

/**************************** GLOBAL VARIABLES *************************/
int     gTCAConfigurationError = TCA_SUCCESS;
char*   gOutputSection;

/**************************** TCA CONFIGURATION FUNCTIONS **************/

/**************************************************************************
 * FUNCTION NAME : ti_tca_reset_egress_tca
 **************************************************************************
 * DESCRIPTION   :
 *  Reset the Egress section of the TCA to the defaults. 
 ***************************************************************************/
static void ti_tca_reset_egress_tca(TI_QOS_TCA* ptr_ti_qos_tca)
{
    int count;

    /* Reset all the egress modes. */
    ptr_ti_qos_tca->tca_mode = ptr_ti_qos_tca->tca_mode & 
                               ~(TI_NSP_QOS_EGRESS_LAYER2 | TI_NSP_QOS_EGRESS_LAYER3);

    /* Initialize the egress mappings. By default all packets are marked with
     * a TOS byte of 0 i.e. treated as a Non-DS Domain. */
    for (count = 0; count < TI_NSP_SERVICE_NUMBER; count++)
        ptr_ti_qos_tca->egress_layer3_mapping[count] = 0x0;

    /* Initialize the egress mappings. By default all packets are marked with 
     * the priority 0. */
    for (count = 0; count < TI_NSP_SERVICE_NUMBER; count++)
        ptr_ti_qos_tca->egress_layer2_mapping[count] = 0x0;

    /* Initialize the defaults for layer 3. */
    ptr_ti_qos_tca->default_non_ip_egress_service_class = TI_NSP_QOS_COS6;
    ptr_ti_qos_tca->default_ip_egress_service_class     = TI_NSP_QOS_COS6;

    /* Initialize the defaults for layer 2. */
    ptr_ti_qos_tca->default_egress_service_class        = TI_NSP_QOS_COS6;
    return;
}

/**************************************************************************
 * FUNCTION NAME : ti_tca_reset_ingress_tca
 **************************************************************************
 * DESCRIPTION   :
 *  Reset the Ingress section of the TCA to the defaults. This function
 *  marks all packets to be low priority. 
 *  
 * NOTES         :
 *  Should PPP Control and ARP packets be treated differently? It might be
 *  useful to push these packets by default as TOP Priority so that even if
 *  the TCA is not configured these packets get preferntial treatment.
 ***************************************************************************/
static void ti_tca_reset_ingress_tca(TI_QOS_TCA* ptr_ti_qos_tca)
{
    int count;

    /* Reset all the ingress modes. By default everything is UNTRUSTED with
     * all packets marked as Low Priority. */
    ptr_ti_qos_tca->tca_mode = ptr_ti_qos_tca->tca_mode & 
                               ~(TI_NSP_QOS_STATIC | TI_NSP_QOS_LAYER3_TRUSTED | TI_NSP_QOS_LAYER2_TRUSTED);

    /* Initialize the mappings. By default all Layer 3 mappings are CoS6 i.e. Best-Effort. */
    for (count = 0; count < MAX_DSCP; count++)
        ptr_ti_qos_tca->ingress_layer3_mapping[count] = TI_NSP_QOS_COS6;

    /* Initialize the mappings. By default all Layer2 mappings are CoS6 i.e. Best-Effort. */
    for (count = 0; count < MAX_VLAN_PRIORITY; count++)
        ptr_ti_qos_tca->ingress_layer2_mapping[count] = TI_NSP_QOS_COS6;

    /* All packets are low priority. */
    ptr_ti_qos_tca->src_interface_service_class = TI_NSP_QOS_COS6;

    /* TODO: Control packets are treated as LOW priority packets by default. */
    ptr_ti_qos_tca->default_non_ip_ingress_service_class = TI_NSP_QOS_COS6;
    return;
}

/**************************************************************************
 * FUNCTION NAME : ti_tcactl_qos_cmd_handler
 **************************************************************************
 * DESCRIPTION   :
 *  This function is the QoS Command Handler.
 *
 * RETURNS       :
 *  -1      - Error.
 *  0       - Success.
 ***************************************************************************/
static int ti_tcactl_qos_cmd_handler(int argc, char* argv[])
{
    /* Validate the number of arguments. */
    if (argc != 2)
    {
        printk ("ERROR: Incorrect format.\n");
        gTCAConfigurationError = TCA_SYNTAX_ERROR;
        return -1;
    }

    /* Validate the arguments. */
    if ((strcmp(argv[1],"enable") != 0) && (strcmp(argv[1],"disable") != 0))
    {
        printk ("ERROR: Invalid values.\n");
        gTCAConfigurationError = TCA_SYNTAX_ERROR;
        return -1;
    }

    /* Set the variable appropriately. */
    if (strcmp(argv[1],"enable") == 0)
        gIsQoSEnabled = 1;
    else
        gIsQoSEnabled = 0;

    /* Return success. The work has been finished. */
    return 0;
}

/**************************************************************************
 * FUNCTION NAME : ti_tcactl_show_cmd_handler
 **************************************************************************
 * DESCRIPTION   :
 *  This function is the Show Command Handler. The function prints the QoS
 *  Framework statistics into the Output Section as well as the console.
 *
 * RETURNS       :
 *  -1      - Error.
 *  0       - Success.
 ***************************************************************************/
static int ti_tcactl_show_cmd_handler(int argc, char* argv[])
{
    int             len = 0;
    TI_QOS_NQM_RR*  ptr_ti_nqm_rr_mcb;
    int             index = 0;

    /* Get the pointer to the Private Information. */
    ptr_ti_nqm_rr_mcb = (TI_QOS_NQM_RR *)ptr_ti_qos_mcb->ptr_nqm_handle->ptr_nqm_private_info;

    /* Print the state of the Framework. */
    len += sprintf (gOutputSection+len, "QOS Framework is %s\n", (gIsQoSEnabled)?"Enabled":"Disabled");

    /* Print the Name of the Scheduling Algorithm. */
    len += sprintf (gOutputSection+len, "Scheduling Algorithm    = %s\n", ptr_ti_qos_mcb->ptr_nqm_handle->name);

    /* Print the Received Statistics. */
    len += sprintf (gOutputSection+len, "-- NQM Received Statistics --\n");
    for (index = 0; index < TI_NSP_SERVICE_NUMBER; index++)
        len += sprintf (gOutputSection+len, "CoS%d Pkts Received = %d\n", index+1, ptr_ti_nqm_rr_mcb->nsp_receive_stats[index]);

    /* Print the Dropped Statistics. */
    len += sprintf (gOutputSection+len, "-- NQM Dropped Statistics --\n");
    for (index = 0; index < TI_NSP_SERVICE_NUMBER; index++)
        len += sprintf (gOutputSection+len, "CoS%d Pkts Dropped  = %d\n", index+1, ptr_ti_nqm_rr_mcb->nsp_dropped_packets[index]);

    /* Print the State of each of the queues. */
    len += sprintf (gOutputSection+len, "-- NQM Congestion Control --\n");
    for (index = 0; index < TI_NSP_SERVICE_NUMBER; index++)
        len += sprintf (gOutputSection+len, "CoS%d Queue = %s\n", index+1, 
                        (ptr_ti_nqm_rr_mcb->nsp_queue_state & (1 << index)) ? "Full" : "Empty");

    /* Print the congestion level */
    if (ptr_ti_nqm_rr_mcb->nsp_congestion_level == TI_QOS_NQM_NOT_CONGESTED)
        len += sprintf (gOutputSection+len, "Congestion State    = Not Congested\n");
    if (ptr_ti_nqm_rr_mcb->nsp_congestion_level == TI_QOS_NQM_DROP_COS1)
        len += sprintf (gOutputSection+len, "Congestion State    = Drop all Packets\n");
    if (ptr_ti_nqm_rr_mcb->nsp_congestion_level == TI_QOS_NQM_DROP_COS2)
        len += sprintf (gOutputSection+len, "Congestion State    = Drop all CoS2,CoS3,CoS4,CoS5,CoS6\n");
    if (ptr_ti_nqm_rr_mcb->nsp_congestion_level == TI_QOS_NQM_DROP_COS3)
        len += sprintf (gOutputSection+len, "Congestion State    = Drop all CoS3,CoS4,CoS5,CoS6\n");
    if (ptr_ti_nqm_rr_mcb->nsp_congestion_level == TI_QOS_NQM_DROP_COS4)
        len += sprintf (gOutputSection+len, "Congestion State    = Drop all CoS4,CoS5,CoS6\n");
    if (ptr_ti_nqm_rr_mcb->nsp_congestion_level == TI_QOS_NQM_DROP_COS5)
        len += sprintf (gOutputSection+len, "Congestion State    = Drop all CoS5,CoS6\n");
    if (ptr_ti_nqm_rr_mcb->nsp_congestion_level == TI_QOS_NQM_DROP_COS6)
        len += sprintf (gOutputSection+len, "Congestion State    = Drop all CoS6\n");

    /* Print the Error counters. */
    len += sprintf (gOutputSection+len, "\n-- Classification Statistics --\n");
    len += sprintf (gOutputSection+len, "Classification Errors   = %d\n", ptr_ti_qos_mcb->classification_error);
    len += sprintf (gOutputSection+len, "Unclassified Packets    = %d\n", ptr_ti_qos_mcb->unclassified_pkts);
    len += sprintf (gOutputSection+len, "Fragmented   Packets    = %d\n", ptr_ti_qos_mcb->fragmented_pkts);

    /* Print the Translation Unit Statistics. */
    len += sprintf(gOutputSection+len,"\n-- Translation Unit Statistics --\n");
    len += sprintf(gOutputSection+len,"Packets Remarked                  = %d\n",ptr_ti_qos_mcb->ptr_tu_handle->tos_byte_remarked);
    len += sprintf(gOutputSection+len,"Packets Unchanged                 = %d\n",ptr_ti_qos_mcb->ptr_tu_handle->pkts_unchanged);
    len += sprintf(gOutputSection+len,"Non-IP Packets Marked             = %d\n",ptr_ti_qos_mcb->ptr_tu_handle->non_ip_pkts);
    len += sprintf(gOutputSection+len,"Unclassified IP Packets Marked    = %d\n",ptr_ti_qos_mcb->ptr_tu_handle->unclassified_ip_pkts);
    len += sprintf(gOutputSection+len,"Unclassified Non-IP Packets Marked= %d\n",ptr_ti_qos_mcb->ptr_tu_handle->unclassified_non_ip_pkts);
    len += sprintf(gOutputSection+len,"Unclassified Layer 2 Packets      = %d\n",ptr_ti_qos_mcb->ptr_tu_handle->unclassified_layer2_pkts);

    /* Sanity Check: Make sure that the Output section has NOT exceeded the buffer size. */
    if (len > MAX_OUTPUT_SECTION_LENGTH)
    {
        /* This is a FATAL Error: We have corrupted the memory now. */
        printk ("FATAL Error: Output section %d bytes written but only %d bytes were allocated.\n", len, MAX_OUTPUT_SECTION_LENGTH);
        gTCAConfigurationError = TCA_INTERNAL_ERROR;
        return -1;
    }

    /* Dump the output section on the console also. */
    printk ("%s", gOutputSection);

    /* The work has been succesfully completed. */
    return 0;
}

/**************************************************************************
 * FUNCTION NAME : ti_tcactl_egress_cmd_handler
 **************************************************************************
 * DESCRIPTION   :
 *  This function is the Egress Command Handler.
 *
 * RETURNS       :
 *  -1      - Error.
 *  0       - Success.
 ***************************************************************************/
static int ti_tcactl_egress_cmd_handler(int argc, char* argv[])
{
    struct net_device*  dev;
    TI_QOS_TCA*         ptr_ti_qos_tca;
    int                 priority = -1;
    char*               command = NULL;
    char*               param1 = NULL;
    char*               ptr_service_class = NULL;
    TI_NSP_QOS_CLASS    service_class = TI_NSP_QOS_COS6;

    /****************************** VALIDATIONS ***************************/

    /* Validate the number of arguments that have been passed. The ingress command should
     * have at least the minimal required set of arguments */
    if (argc < 3)
    {
        printk("ERROR: Incorrect Number of parameters passed.\n");
        gTCAConfigurationError = TCA_SYNTAX_ERROR;
        return -1;
    }

    /* Validate if the device passed is actually present or not. */ 
	dev = dev_get_by_name(argv[1]);
    if (dev == NULL)
    {
        printk ("ERROR: Device %s does not exist.\n", argv[1]);
        gTCAConfigurationError = TCA_INTERNAL_ERROR;
        return -1;
    }

    /* Release the hold on the device. */
    dev_put (dev);

    /* Remember the command that has been passed. */
    command = argv[2];

    /* Validate the command parameters. */
    if ((strcmp (command, "layer2") != 0) && (strcmp (command, "layer3") != 0) &&
        (strcmp (command, "show") != 0) && (strcmp (command, "reset") != 0) &&
        (strcmp (command, "layer2default") != 0) && (strcmp (command, "layer3default-non-ip") != 0) &&
        (strcmp (command, "layer3default-ip") != 0))
    {
        printk ("ERROR: Incorrect command %s.\n", command);
        gTCAConfigurationError = TCA_SYNTAX_ERROR;
        return -1;
    }

    /* Validate the layer2 command. */
    if (strcmp (command, "layer2") == 0)
    {
        /* Validate the number of arguments. */
        if (argc != 5)
        {
            printk ("ERROR: Layer 2 Command has incorrect number of arguments.\n");
            gTCAConfigurationError = TCA_SYNTAX_ERROR;
            return -1;
        }

        /* Get the additional parameters */
        ptr_service_class   = argv[3];
        param1              = argv[4];

        /* Get the priority value that has been passed. */
        priority = (int) simple_strtol(param1, NULL, 0);
        if ((priority < 0) || (priority > MAX_VLAN_PRIORITY))
        {
            printk ("ERROR: Incorrect User Priority Value %d should be within 0-7.\n",priority);
            gTCAConfigurationError = TCA_SYNTAX_ERROR;
            return -1;
        }
    }

    /* Validate the layer3 command. */
    if (strcmp (command, "layer3") == 0)
    {
        /* Validate the number of arguments. */
        if (argc != 5)
        {
            printk ("ERROR: Layer 3 Command has incorrect number of arguments.\n");
            gTCAConfigurationError = TCA_SYNTAX_ERROR;
            return -1;
        }

        /* Get the additional parameters */
        ptr_service_class   = argv[3];
        param1              = argv[4];

        /* Get the priority value that has been passed. */
        priority = (int) simple_strtol(param1, NULL, 0);
        if ((priority < 0) || (priority > MAX_DSCP))
        {
            printk ("ERROR: Incorrect TOS Byte Value %d should be within 0-255.\n",priority);
            gTCAConfigurationError = TCA_SYNTAX_ERROR;
            return -1;
        }
    }

    /* Validate the layer3default-non-ip command. */
    if (strcmp (command, "layer3default-non-ip") == 0)
    {
        /* Validate the number of arguments. */
        if (argc != 4)
        {
            printk ("ERROR: layer3default-non-ip has incorrect number of arguments.\n");
            gTCAConfigurationError = TCA_SYNTAX_ERROR;
            return -1;
        }

        /* Get the additional parameters */
        ptr_service_class = argv[3];
    }

    /* Validate the layer3default-non-ip command. */
    if (strcmp (command, "layer3default-ip") == 0)
    {
        /* Validate the number of arguments. */
        if (argc != 4)
        {
            printk ("ERROR: layer3default-ip has incorrect number of arguments.\n");
            gTCAConfigurationError = TCA_SYNTAX_ERROR;
            return -1;
        }

        /* Get the additional parameters */
        ptr_service_class = argv[3];
    }

    /* Validate the layer2default command. */
    if (strcmp (command, "layer2default") == 0)
    {
        /* Validate the number of arguments. */
        if (argc != 4)
        {
            printk ("ERROR: layer2default has incorrect number of arguments.\n");
            gTCAConfigurationError = TCA_SYNTAX_ERROR;
            return -1;
        }

        /* Get the additional parameters */
        ptr_service_class = argv[3];
    }

    /* Validate the service class if passed. */
    if (ptr_service_class != NULL)
    {
        if ((strcmp (ptr_service_class, "CoS1") != 0) && (strcmp (ptr_service_class, "CoS2") != 0) &&
            (strcmp (ptr_service_class, "CoS3") != 0) && (strcmp (ptr_service_class, "CoS4") != 0) &&
            (strcmp (ptr_service_class, "CoS5") != 0) && (strcmp (ptr_service_class, "CoS6") != 0))
        {
            printk ("ERROR: Incorrect Class of Service parameter should be CoS1 - CoS6 but was %s.\n", ptr_service_class);
            gTCAConfigurationError = TCA_SYNTAX_ERROR;
            return -1;
        }

        /* Get the actual service class. */
        if (strcmp (ptr_service_class, "CoS1") == 0)
            service_class = TI_NSP_QOS_COS1;
        else if (strcmp (ptr_service_class, "CoS2") == 0)
            service_class = TI_NSP_QOS_COS2;
        else if (strcmp (ptr_service_class, "CoS3") == 0)
            service_class = TI_NSP_QOS_COS3;
        else if (strcmp (ptr_service_class, "CoS4") == 0)
            service_class = TI_NSP_QOS_COS4;
        else if (strcmp (ptr_service_class, "CoS5") == 0)
            service_class = TI_NSP_QOS_COS5;
        else
            service_class = TI_NSP_QOS_COS6;
    }

    /************************ End of VALIDATIONS ******************************/

    /* There should always be a TCA associated with the device. */
    if (dev->ptr_ti_qos_tca == NULL)
    {
        printk ("DEBUG: No TCA for device %s creating default\n", dev->name);

        /* There was no TCA associated with the device. This could be a case
         * for the loopback device; so lets create one. */
        if (ti_tca_device_init (dev) < 0)
        {
            gTCAConfigurationError = TCA_INTERNAL_ERROR;
            return -1;
        }
    }

    /* Initialize the Pointer to the TCA. */
    ptr_ti_qos_tca = dev->ptr_ti_qos_tca;

    /************************ Command Processing ******************************/

    /* Check if the command was "reset" */
    if (strcmp(command, "reset") == 0)
    {
        /* Reset the Ingress TCA to the defaults. */
        ti_tca_reset_egress_tca (ptr_ti_qos_tca);
        return 0;        
    }

    /* Check if the command was "show" */
    if (strcmp (command, "show") == 0)
    {
        int i;
        int len = 0;

        /* There was a TCA associated with the device. Print out all the information to the output
         * section and to the console simulatenously. */
        if (ptr_ti_qos_tca->tca_mode & TI_NSP_QOS_EGRESS_LAYER2)
        {
            /* Layer 2 Trusted Mode. */
            len += sprintf (gOutputSection+len, "TCA Mode: Layer2 Egress\n");

            /* Print the mappings. */
            for (i = 0; i < TI_NSP_SERVICE_NUMBER; i++)
                len += sprintf (gOutputSection+len, "CoS%d=%d.\n",i+1, ptr_ti_qos_tca->egress_layer2_mapping[i]);

            /* Print the default mappings for the unclassified packets. */
            len += sprintf (gOutputSection+len, "Unclassified Packets=CoS%d.\n",ptr_ti_qos_tca->default_egress_service_class+1);
        }
        else 
        {
            if (ptr_ti_qos_tca->tca_mode & TI_NSP_QOS_EGRESS_LAYER3)
            {
                /* Layer 3 Trusted Mode. */
                len += sprintf (gOutputSection+len, "TCA Mode: Layer3 Egress\n");

                /* Print the mappings. */
                for (i = 0; i < TI_NSP_SERVICE_NUMBER; i++)
                    len += sprintf (gOutputSection+len,"CoS%d=%d.\n", i+1, ptr_ti_qos_tca->egress_layer3_mapping[i]);

                /* Print the defaults mappings for unclassified IP packets. */
                len += sprintf (gOutputSection+len,"Unclassified IP Packets=CoS%d.\n", 
                                ptr_ti_qos_tca->default_ip_egress_service_class+1);

                /* Print the defaults mappings for unclassified non-IP packets. */
                len += sprintf (gOutputSection+len,"Unclassified Non-IP Packets=CoS%d.\n",
                                ptr_ti_qos_tca->default_non_ip_egress_service_class+1);
            }
            else
            {
                /* There is NO TCA defined. */
                sprintf (gOutputSection+len,"TCA Mode: No Egress TCA defined.\n");
            }
        }

        /* Sanity Check: Make sure that the Output section has NOT exceeded
         * the buffer size. */
        if (len > MAX_OUTPUT_SECTION_LENGTH)
        {
            /* This is a FATAL Error: We have corrupted the memory now. */
            printk ("FATAL Error: Output section %d bytes written but only %d bytes were allocated.\n", 
                    len, MAX_OUTPUT_SECTION_LENGTH);
            gTCAConfigurationError = TCA_INTERNAL_ERROR;
            return -1;
        }

        /* Print the output section on the console. */
        printk ("%s\n", gOutputSection);
        return 0;
    }

    /* Reset all the egress modes.. Only 1 is valid. */
    ptr_ti_qos_tca->tca_mode = ptr_ti_qos_tca->tca_mode & 
                               ~(TI_NSP_QOS_EGRESS_LAYER2 | TI_NSP_QOS_EGRESS_LAYER3);

    /* Check if the command was "layer2" */
    if (strcmp (command, "layer2") == 0)
    {
        /* Set the TCA mode to operate at layer2. */
        ptr_ti_qos_tca->tca_mode = ptr_ti_qos_tca->tca_mode | TI_NSP_QOS_EGRESS_LAYER2;
        ptr_ti_qos_tca->egress_layer2_mapping[service_class] = priority;
        return 0;
    }

    /* Check if the command was "layer2default" */
    if (strcmp (command, "layer2default") == 0)
    {
        /* Set the TCA mode to operate at layer2. */
        ptr_ti_qos_tca->tca_mode = ptr_ti_qos_tca->tca_mode | TI_NSP_QOS_EGRESS_LAYER2;
        ptr_ti_qos_tca->default_egress_service_class = service_class;
        return 0;
    }

    /* Check if the command was "layer3" */
    if (strcmp (command, "layer3") == 0)
    {
        /* Set the TCA mode to operate at layer2. */
        ptr_ti_qos_tca->tca_mode = ptr_ti_qos_tca->tca_mode | TI_NSP_QOS_EGRESS_LAYER3;
        ptr_ti_qos_tca->egress_layer3_mapping[service_class] = priority;
        return 0;
    }

    /* Check if the command was "layer3default-non-ip" */
    if (strcmp (command, "layer3default-non-ip") == 0)
    {
        /* Set the TCA mode to operate at layer2. */
        ptr_ti_qos_tca->tca_mode = ptr_ti_qos_tca->tca_mode | TI_NSP_QOS_EGRESS_LAYER3;
        ptr_ti_qos_tca->default_non_ip_egress_service_class = service_class;
        return 0;
    }

    /* Check if the command was "layer3default-non-ip" */
    if (strcmp (command, "layer3default-ip") == 0)
    {
        /* Set the TCA mode to operate at layer2. */
        ptr_ti_qos_tca->tca_mode = ptr_ti_qos_tca->tca_mode | TI_NSP_QOS_EGRESS_LAYER3;
        ptr_ti_qos_tca->default_ip_egress_service_class = service_class;
        return 0;
    }

    /* Code should never come here. */
    gTCAConfigurationError = TCA_INTERNAL_ERROR;
    return -1;
}

/**************************************************************************
 * FUNCTION NAME : ti_tcactl_ingress_cmd_handler
 **************************************************************************
 * DESCRIPTION   :
 *  This function is the Ingress Command Handler.
 *
 * RETURNS       :
 *  -1      - Error.
 *  0       - Success.
 ***************************************************************************/
static int ti_tcactl_ingress_cmd_handler(int argc, char* argv[])
{
    struct net_device*  dev;
    TI_QOS_TCA*         ptr_ti_qos_tca;
    int                 priority = -1;
    char*               command = NULL;
    char*               param1 = NULL;
    char*               ptr_service_class = NULL;
    TI_NSP_QOS_CLASS    service_class = TI_NSP_QOS_COS6;
 
    /****************************** VALIDATIONS ***************************/

    /* Validate the number of arguments that have been passed. The ingress command should
     * have at least the minimal required set of arguments */
    if (argc < 3)
    {
        printk ("ERROR: Incorrect Number of parameters passed.\n");
        gTCAConfigurationError = TCA_SYNTAX_ERROR;
        return -1;
    }

    /* Validate if the device passed is actually present or not. */ 
	dev = dev_get_by_name(argv[1]);
    if (dev == NULL)
    {
        printk ("ERROR: Device %s does not exist.\n", argv[1]);
        gTCAConfigurationError = TCA_INTERNAL_ERROR;
        return -1;
    }

    /* Release the hold on the device. */
    dev_put (dev);

    /* Remember the command that has been passed. */
    command = argv[2];

    /* Validate the command parameters. */
    if ((strcmp (command, "layer2trusted") != 0) && (strcmp (command, "layer3trusted") != 0) &&
        (strcmp (command, "show") != 0) && (strcmp (command, "reset") != 0) &&
        (strcmp (command, "static") != 0))
    {
        printk ("ERROR: Incorrect command %s.\n", command);
        gTCAConfigurationError = TCA_SYNTAX_ERROR;
        return -1;
    }

    /* Validate the layer2trusted command. */
    if (strcmp (command, "layer2trusted") == 0)
    {
        /* Validate the number of arguments. */
        if (argc != 5)
        {
            printk ("ERROR: Layer 2 Trusted Commands have incorrect number of arguments.\n");
            gTCAConfigurationError = TCA_SYNTAX_ERROR;
            return -1;
        }

        /* Get the additional parameters */
        param1              = argv[3];
        ptr_service_class   = argv[4];

        /* Get the priority value that has been passed. */
        priority = (int) simple_strtol(param1, NULL, 0);
        if ((priority < 0) || (priority > MAX_VLAN_PRIORITY))
        {
            printk ("ERROR: Incorrect User Priority Value %d should be within 0-7.\n",priority);
            gTCAConfigurationError = TCA_SYNTAX_ERROR;
            return -1;
        }
    }

    /* Validate the layer3trusted command. */
    if (strcmp (command, "layer3trusted") == 0)
    {
        /* Validate the number of arguments. */
        if (argc != 5)
        {
            printk ("ERROR: Layer 3 Trusted Commands have incorrect number of arguments.\n");
            gTCAConfigurationError = TCA_SYNTAX_ERROR;
            return -1;
        }

        /* Get the additional parameters */
        param1              = argv[3];
        ptr_service_class   = argv[4];

        /* The first parameter could be a default value. */
        if (strcmp(param1, "default-non-ip") != 0)
        {
            /* Get the priority value that has been passed. */
            priority = (int) simple_strtol(param1, NULL, 0);
            if ((priority < 0) || (priority > MAX_DSCP))
            {
                printk ("ERROR: Incorrect TOS Value %d should be within 0-255.\n",priority);
                gTCAConfigurationError = TCA_SYNTAX_ERROR;
                return -1;
            }
        }
    }

    /* Validate the static command. */
    if (strcmp (command, "static") == 0)
    {
        /* Validate the number of arguments. */
        if (argc != 4)
        {
            printk ("ERROR: Static Command has an incorrect number of arguments.\n");
            gTCAConfigurationError = TCA_SYNTAX_ERROR;
            return -1;
        }

        /* Get the service class associated. */
        ptr_service_class = argv[3];
    }

    /* Validate the service class if passed. */
    if (ptr_service_class != NULL)
    {
        if ((strcmp (ptr_service_class, "CoS1") != 0) && (strcmp (ptr_service_class, "CoS2") != 0) &&
            (strcmp (ptr_service_class, "CoS3") != 0) && (strcmp (ptr_service_class, "CoS4") != 0) &&
            (strcmp (ptr_service_class, "CoS5") != 0) && (strcmp (ptr_service_class, "CoS6") != 0))
        {
            printk ("ERROR: Incorrect Class of Service parameter should be CoS1 - CoS6 but was %s.\n", ptr_service_class);
            gTCAConfigurationError = TCA_SYNTAX_ERROR;
            return -1;
        }

        /* Get the actual service class. */
        if (strcmp (ptr_service_class, "CoS1") == 0)
            service_class = TI_NSP_QOS_COS1;
        else if (strcmp (ptr_service_class, "CoS2") == 0)
            service_class = TI_NSP_QOS_COS2;
        else if (strcmp (ptr_service_class, "CoS3") == 0)
            service_class = TI_NSP_QOS_COS3;
        else if (strcmp (ptr_service_class, "CoS4") == 0)
            service_class = TI_NSP_QOS_COS4;
        else if (strcmp (ptr_service_class, "CoS5") == 0)
            service_class = TI_NSP_QOS_COS5;
        else
            service_class = TI_NSP_QOS_COS6;
    }

    /************************ End of VALIDATIONS ******************************/

    /* There should always be a TCA associated with the device. */
    if (dev->ptr_ti_qos_tca == NULL)
    {
        printk ("DEBUG: No TCA for device %s creating default\n", dev->name);

        /* There was no TCA associated with the device. This could be a case
         * for the loopback device; so lets create one. */
        if (ti_tca_device_init (dev) < 0)
        {
            gTCAConfigurationError = TCA_INTERNAL_ERROR;
            return -1;
        }
    }

    /* Initialize the Pointer to the TCA. */
    ptr_ti_qos_tca = dev->ptr_ti_qos_tca;

    /************************ Command Processing ******************************/

    /* Check if the command was "reset" */
    if (strcmp(command, "reset") == 0)
    {
        /* Reset the Ingress TCA to the defaults. */
        ti_tca_reset_ingress_tca (ptr_ti_qos_tca);
        return 0;        
    }

    /* Check if the command was "show" */
    if (strcmp (command, "show") == 0)
    {
        int i;
        int len = 0;

        /* There was a TCA associated with the device. Print out all the information on the console
         * and in the output section. */
        if (ptr_ti_qos_tca->tca_mode & TI_NSP_QOS_LAYER2_TRUSTED)
        {
            /* Layer 2 Trusted Mode. */
            len += sprintf (gOutputSection+len, "TCA Mode: Layer2 Trusted\n");

            /* Print the mappings */
            for (i = 0; i < MAX_VLAN_PRIORITY; i++)
                len += sprintf (gOutputSection+len, "%d=CoS%d.\n", i, ptr_ti_qos_tca->ingress_layer2_mapping[i]+1);
        }
        else 
        {
            if (ptr_ti_qos_tca->tca_mode & TI_NSP_QOS_LAYER3_TRUSTED)
            {
                /* Layer 3 Trusted Mode. */
                len += sprintf (gOutputSection+len, "TCA Mode: Layer3 Trusted\n");

                /* Print the mappings but do so only for non CoS6 mappings. */
                for (i = 0; i < MAX_DSCP; i++)
                {
                    if (ptr_ti_qos_tca->ingress_layer3_mapping[i] != TI_NSP_QOS_COS6)
                        len += sprintf (gOutputSection+len, "%d=CoS%d.\n", i, ptr_ti_qos_tca->ingress_layer3_mapping[i]+1);
                }

                /* Print the default classifications for non-ip packets. */
                len += sprintf (gOutputSection+len,"NonIP=CoS%d.\n",ptr_ti_qos_tca->default_non_ip_ingress_service_class+1);
            }
            else
            {
                if (ptr_ti_qos_tca->tca_mode & TI_NSP_QOS_STATIC)
                {
                    /* Static Mode. */
                    len += sprintf (gOutputSection+len, "TCA Mode: Static Mode\n");

                    /* Print the mappings. */
                    len += sprintf (gOutputSection+len,"All packets=CoS%d.\n", ptr_ti_qos_tca->src_interface_service_class+1);
                }
                else
                {
                    /* Untrusted Mode. */
                    len += sprintf (gOutputSection+len, "TCA Mode: Untrusted Mode\n");
                }
            }
        }

        /* Sanity Check: Make sure that the Output section has NOT exceeded
         * the buffer size. */
        if (len > MAX_OUTPUT_SECTION_LENGTH)
        {
            /* This is a FATAL Error: We have corrupted the memory now. */
            printk ("FATAL Error: Output section %d bytes written but only %d bytes were allocated.\n", 
                    len, MAX_OUTPUT_SECTION_LENGTH);
            gTCAConfigurationError = TCA_INTERNAL_ERROR;
            return -1;
        }

        /* Print the output section on the console. */
        printk ("%s\n", gOutputSection);
        return 0;
    }

    /* Reset all the ingress modes.. Only 1 is valid. */
    ptr_ti_qos_tca->tca_mode = ptr_ti_qos_tca->tca_mode & 
                               ~(TI_NSP_QOS_STATIC | TI_NSP_QOS_LAYER3_TRUSTED | TI_NSP_QOS_LAYER2_TRUSTED);

    /* Check if the command was "static" */
    if (strcmp (command, "static") == 0)
    {
        /* Static Mode. */
        ptr_ti_qos_tca->tca_mode = ptr_ti_qos_tca->tca_mode | TI_NSP_QOS_STATIC;
        ptr_ti_qos_tca->src_interface_service_class = service_class;
        return 0;
    }

    /* Check if the command was "layer2trusted" */
    if (strcmp (command, "layer2trusted") == 0)
    {
        ptr_ti_qos_tca->tca_mode = ptr_ti_qos_tca->tca_mode | TI_NSP_QOS_LAYER2_TRUSTED;
        ptr_ti_qos_tca->ingress_layer2_mapping[priority] = service_class;
        return 0;
    }

    /* Check if the command was "layer3trusted" */
    if (strcmp (command, "layer3trusted") == 0)
    {
        ptr_ti_qos_tca->tca_mode = ptr_ti_qos_tca->tca_mode | TI_NSP_QOS_LAYER3_TRUSTED;

        /* The first parameter could be a default value. */
        if (strcmp(param1, "default-non-ip") == 0)
            ptr_ti_qos_tca->default_non_ip_ingress_service_class = service_class;
        else
           ptr_ti_qos_tca->ingress_layer3_mapping[priority] = service_class;

        return 0;
    }

    /* Code should never reach here. */
    return -1;
}

/**************************************************************************
 * FUNCTION NAME : ti_tcactl_show_stats
 **************************************************************************
 * DESCRIPTION   :
 *  The function prints the QoS Internal Statistics.
 ***************************************************************************/
static void ti_tcactl_show_stats(void)
{
    TI_QOS_NQM_RR*  ptr_ti_nqm_rr_mcb;
    int             index = 0;

    /* Get the pointer to the Private Information. */
    ptr_ti_nqm_rr_mcb = (TI_QOS_NQM_RR *)ptr_ti_qos_mcb->ptr_nqm_handle->ptr_nqm_private_info;

    printk("QOS Framework is %s\n", (gIsQoSEnabled)?"Enabled":"Disabled");

    printk("********** Network Queue Manager Statistics **********\n");

    /* Print the Name of the Scheduling Algorithm. */
    printk( "Scheduling Algorithm    = %s\n", ptr_ti_qos_mcb->ptr_nqm_handle->name);

    /* Print the Received Statistics. */
    printk ("------ NQM Received Statistics ------\n");
    for (index = 0; index < TI_NSP_SERVICE_NUMBER; index++)
        printk ("CoS%d Pkts Received = %d\n", index+1, ptr_ti_nqm_rr_mcb->nsp_receive_stats[index]);

    /* Print the Dropped Statistics. */
    printk ("------ NQM Dropped Statistics ------\n");
    for (index = 0; index < TI_NSP_SERVICE_NUMBER; index++)
        printk ("CoS%d Pkts Dropped  = %d\n", index+1, ptr_ti_nqm_rr_mcb->nsp_dropped_packets[index]);

    /* Print the State of each of the queues. */
    printk ("------ NQM Congestion Control ------\n");
    for (index = 0; index < TI_NSP_SERVICE_NUMBER; index++)
        printk ("CoS%d Queue = %s\n", index+1, (ptr_ti_nqm_rr_mcb->nsp_queue_state & (1 << index)) ? "Full" : "Empty");

    /* Print the congestion level */
    if (ptr_ti_nqm_rr_mcb->nsp_congestion_level == TI_QOS_NQM_NOT_CONGESTED)
        printk ("Congestion State    = Not Congested\n");
    if (ptr_ti_nqm_rr_mcb->nsp_congestion_level == TI_QOS_NQM_DROP_COS1)
        printk ("Congestion State    = Drop all Packets\n");
    if (ptr_ti_nqm_rr_mcb->nsp_congestion_level == TI_QOS_NQM_DROP_COS2)
        printk ("Congestion State    = Drop all CoS2,CoS3,CoS4,CoS5,CoS6\n");
    if (ptr_ti_nqm_rr_mcb->nsp_congestion_level == TI_QOS_NQM_DROP_COS3)
        printk ("Congestion State    = Drop all CoS3,CoS4,CoS5,CoS6\n");
    if (ptr_ti_nqm_rr_mcb->nsp_congestion_level == TI_QOS_NQM_DROP_COS4)
        printk ("Congestion State    = Drop all CoS4,CoS5,CoS6\n");
    if (ptr_ti_nqm_rr_mcb->nsp_congestion_level == TI_QOS_NQM_DROP_COS5)
        printk ("Congestion State    = Drop all CoS5,CoS6\n");
    if (ptr_ti_nqm_rr_mcb->nsp_congestion_level == TI_QOS_NQM_DROP_COS6)
        printk ("Congestion State    = Drop all CoS6\n");

    /* Print the Error counters. */
    printk ("\n********** Classification Statistics **********\n");
    printk ("Classification Errors   = %d\n", ptr_ti_qos_mcb->classification_error);
    printk ("Unclassified Packets    = %d\n", ptr_ti_qos_mcb->unclassified_pkts);
    printk ("Fragmented   Packets    = %d\n", ptr_ti_qos_mcb->fragmented_pkts);

    /* Print the Translation Unit Statistics. */
    printk ("\n********** Translation Unit Statistics **********\n");
    printk ("Packets Remarked                  = %d\n", ptr_ti_qos_mcb->ptr_tu_handle->tos_byte_remarked);
    printk ("Packets Unchanged                 = %d\n", ptr_ti_qos_mcb->ptr_tu_handle->pkts_unchanged);
    printk ("Non-IP Packets Marked             = %d\n", ptr_ti_qos_mcb->ptr_tu_handle->non_ip_pkts);
    printk ("Unclassified IP Packets Marked    = %d\n", ptr_ti_qos_mcb->ptr_tu_handle->unclassified_ip_pkts);
    printk ("Unclassified Non-IP Packets Marked= %d\n", ptr_ti_qos_mcb->ptr_tu_handle->unclassified_non_ip_pkts);
    printk ("Unclassified Layer 2 Packets      = %d\n", ptr_ti_qos_mcb->ptr_tu_handle->unclassified_layer2_pkts);

    return;
}

/**************************************************************************
 * FUNCTION NAME : ti_tcactl_write_info
 **************************************************************************
 * DESCRIPTION   :
 *  This function is the heart of the TCA Configuration Engine. All the
 *  commands are passed to this function. The function parses the commands
 *  and then passes them to the appropriate handlers.
 *
 * RETURNS       :
 *  -1              - Error.
 *  Non-Zero        - Success.
 ***************************************************************************/
static int ti_tcactl_write_info (struct file *file, const char *buffer, unsigned long count, void *data) 
{
	char    qos[100];
	char*	argv[10];
	int		argc = 0;
	char*	ptr_cmd;
    char*   delimitters = " \n\t";

	/* Validate the length of data passed. */
	if (count > 100)
		count = 100; 

    /* Initialize the buffer before using it. */
    memset ((void *)&qos[0], 0, sizeof(qos));
    memset ((void *)&argv[0], 0, sizeof(argv));

    /* Initialize the output section to be empty. */
    memset ((void *)gOutputSection, 0, MAX_OUTPUT_SECTION_LENGTH);

    /* Initialize the status variable to success. */
    gTCAConfigurationError = TCA_SUCCESS;
	
	/* Copy from user space. */	
	if (copy_from_user (&qos, buffer, count))
    {
        gTCAConfigurationError = TCA_INTERNAL_ERROR;
		return -EFAULT;
    }

	/* Tokenize the command. Check if there was a NULL entry. If so be the case the
     * user did not know how to use the entry. Print the help screen. */
	ptr_cmd = strtok (qos, delimitters);
    if (ptr_cmd == NULL)
    {
        gTCAConfigurationError = TCA_INTERNAL_ERROR;
        return count;
    }

    /* Parse all the commands typed. */
	do
	{
        /* Extract the first command. */
		argv[argc++] = ptr_cmd;

        /* Validate if the user entered more commands.*/
		if (argc >=10)
		{
			printk ("ERROR: Incorrect too many parameters dropping the command\n");
            gTCAConfigurationError = TCA_INTERNAL_ERROR;
			return -EFAULT;
		}

        /* Get the next valid command. */
		ptr_cmd = strtok (NULL, delimitters);
	} while (ptr_cmd != NULL);

#if 0
    /* DEBUG TEST CODE -----*/
    {
        int i = 0;
        printk ("DEBUG: Detected %d arguments.\n", argc);
        for (i = 0; i < argc; i++)
           printk ("DEBUG: Argument %d = %s.\n", i, argv[i]);
    }
#endif

    /******************************* Command Handlers *******************************/

    /* Ingress Command Handlers */
    if (strncmp(argv[0], "ingress", strlen("ingress")) == 0)
    {
        /* Call the Ingress Command Handler. */
        if (ti_tcactl_ingress_cmd_handler (argc, argv) < 0)
            return -1;

        return count;
    }

    /* Egress Command Handlers */
    if (strncmp(argv[0], "egress", strlen("egress")) == 0)
    {
        /* Call the Egress Command Handler. */
        if (ti_tcactl_egress_cmd_handler (argc, argv) < 0)
            return -1;

        return count;
    }

    /* QoS Framework Handler. */
    if (strncmp(argv[0], "qos", strlen("qos")) == 0)
    {
        /* Call the QoS Command Handler. */
        if (ti_tcactl_qos_cmd_handler(argc, argv) < 0)
            return -1;

        return count;
    }

    /* Show Command Handler. */
    if (strncmp(argv[0], "show", strlen("show")) == 0)
    {
        /* Call the QoS Command Handler. */
        if (ti_tcactl_show_cmd_handler(argc, argv) < 0)
            return -1;

        return count;
    }

    /* ERROR: Invalid command. */
    gTCAConfigurationError = TCA_INTERNAL_ERROR;
    return -1;
}

/**************************************************************************
 * FUNCTION NAME : ti_tcactl_read_info
 **************************************************************************
 * DESCRIPTION   :
 *  This entry is used in conjunction with the write block and has a special
 *  use that is used to convey the status of the previous write operation.
 ***************************************************************************/
static int ti_tcactl_read_info (char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
    /* Place the global status variable into the buffer. */
    int len = 0;
    len += sprintf (buf+len, "Return Value: %d\n", gTCAConfigurationError);
    len += sprintf (buf+len, "Output: \n");

    /* Print the output section. */
    len += sprintf (buf+len, "%s", gOutputSection);
    return len;
}

#if 0
/**************************************************************************
 * FUNCTION NAME : ti_tcactl_show_qos_stats
 **************************************************************************
 * DESCRIPTION   :
 *  The function is the READ Entry for the PROC Entry that dumps the internal
 *  statistics reported by the various internal QoS Functional Blocks.
 *
 * RETURNS       :
 *  Length of the buffer that has all the statistical information.
 ***************************************************************************/
static int ti_tcactl_show_qos_stats(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
    int             len = 0;
    TI_QOS_NQM_RR*  ptr_ti_nqm_rr_mcb;
    int             index = 0;

    /* Get the pointer to the Private Information. */
    ptr_ti_nqm_rr_mcb = (TI_QOS_NQM_RR *)ptr_ti_qos_mcb->ptr_nqm_handle->ptr_nqm_private_info;

    len += sprintf (buf+len, "QOS Framework is %s\n", (gIsQoSEnabled)?"Enabled":"Disabled");

    len += sprintf (buf+len, "********** Network Queue Manager Statistics **********\n");

    /* Print the Name of the Scheduling Algorithm. */
    len += sprintf (buf+len, "Scheduling Algorithm    = %s\n", ptr_ti_qos_mcb->ptr_nqm_handle->name);

    /* Print the Received Statistics. */
    len += sprintf (buf+len, "------ NQM Received Statistics ------\n");
    for (index = 0; index < TI_NSP_SERVICE_NUMBER; index++)
        len += sprintf (buf+len, "CoS%d Pkts Received = %d\n", index+1, ptr_ti_nqm_rr_mcb->nsp_receive_stats[index]);

    /* Print the Dropped Statistics. */
    len += sprintf (buf+len, "------ NQM Dropped Statistics ------\n");
    for (index = 0; index < TI_NSP_SERVICE_NUMBER; index++)
        len += sprintf (buf+len, "CoS%d Pkts Dropped  = %d\n", index+1, ptr_ti_nqm_rr_mcb->nsp_dropped_packets[index]);

    /* Print the State of each of the queues. */
    len += sprintf (buf+len, "------ NQM Congestion Control ------\n");
    for (index = 0; index < TI_NSP_SERVICE_NUMBER; index++)
        len += sprintf (buf+len, "CoS%d Queue = %s\n", index+1, 
                        (ptr_ti_nqm_rr_mcb->nsp_queue_state & (1 << index)) ? "Full" : "Empty");

    /* Print the congestion level */
    if (ptr_ti_nqm_rr_mcb->nsp_congestion_level == TI_QOS_NQM_NOT_CONGESTED)
        len += sprintf (buf+len, "Congestion State    = Not Congested\n");
    if (ptr_ti_nqm_rr_mcb->nsp_congestion_level == TI_QOS_NQM_DROP_COS1)
        len += sprintf (buf+len, "Congestion State    = Drop all Packets\n");
    if (ptr_ti_nqm_rr_mcb->nsp_congestion_level == TI_QOS_NQM_DROP_COS2)
        len += sprintf (buf+len, "Congestion State    = Drop all CoS2,CoS3,CoS4,CoS5,CoS6\n");
    if (ptr_ti_nqm_rr_mcb->nsp_congestion_level == TI_QOS_NQM_DROP_COS3)
        len += sprintf (buf+len, "Congestion State    = Drop all CoS3,CoS4,CoS5,CoS6\n");
    if (ptr_ti_nqm_rr_mcb->nsp_congestion_level == TI_QOS_NQM_DROP_COS4)
        len += sprintf (buf+len, "Congestion State    = Drop all CoS4,CoS5,CoS6\n");
    if (ptr_ti_nqm_rr_mcb->nsp_congestion_level == TI_QOS_NQM_DROP_COS5)
        len += sprintf (buf+len, "Congestion State    = Drop all CoS5,CoS6\n");
    if (ptr_ti_nqm_rr_mcb->nsp_congestion_level == TI_QOS_NQM_DROP_COS6)
        len += sprintf (buf+len, "Congestion State    = Drop all CoS6\n");

    /* Print the Error counters. */
    len += sprintf (buf+len, "\n********** Classification Statistics **********\n");
    len += sprintf (buf+len, "Classification Errors   = %d\n", ptr_ti_qos_mcb->classification_error);
    len += sprintf (buf+len, "Unclassified Packets    = %d\n", ptr_ti_qos_mcb->unclassified_pkts);

    /* Print the Translation Unit Statistics. */
    len += sprintf (buf+len, "\n********** Translation Unit Statistics **********\n");
    len += sprintf (buf+len, "Packets Remarked                  = %d\n", ptr_ti_qos_mcb->ptr_tu_handle->tos_byte_remarked);
    len += sprintf (buf+len, "Packets Unchanged                 = %d\n", ptr_ti_qos_mcb->ptr_tu_handle->pkts_unchanged);
    len += sprintf (buf+len, "Non-IP Packets Marked             = %d\n", ptr_ti_qos_mcb->ptr_tu_handle->non_ip_pkts);
    len += sprintf (buf+len, "Unclassified IP Packets Marked    = %d\n", ptr_ti_qos_mcb->ptr_tu_handle->unclassified_ip_pkts);
    len += sprintf (buf+len, "Unclassified Non-IP Packets Marked= %d\n", ptr_ti_qos_mcb->ptr_tu_handle->unclassified_non_ip_pkts);
    len += sprintf (buf+len, "Unclassified Layer 2 Packets      = %d\n", ptr_ti_qos_mcb->ptr_tu_handle->unclassified_layer2_pkts);

    return len;
}
#endif

/**************************************************************************
 * FUNCTION NAME : ti_nsp_qos_init_round_robin_scheduler
 **************************************************************************
 * DESCRIPTION   :
 *  The function initializes the Round-Robin scheduler internals and hooks the
 *  NQM with it. 
 ***************************************************************************/
static int ti_nsp_qos_init_round_robin_scheduler(TI_QOS_NQM_MCB* ptr_ti_qos_nqm_mcb)
{
    TI_QOS_NQM_RR* ptr_nqm_qos_rr;

    /* Allocate memory for the Round-Robin Scheduler and initialize the memory block. */
    ptr_nqm_qos_rr = (TI_QOS_NQM_RR *)kmalloc(sizeof(TI_QOS_NQM_RR), GFP_KERNEL);
    if (ptr_nqm_qos_rr == NULL)
    {
        printk ("Error: No memory to allocate the Round-Robin Scheduler.\n");
        return -1;
    }
    memset ((void *)ptr_nqm_qos_rr, 0, sizeof(TI_QOS_NQM_RR));

    /* Fill up the NQM QoS Round-Robin Structure. */
    ptr_ti_qos_nqm_mcb->name                  = "Strict Round-Robin";
    ptr_ti_qos_nqm_mcb->enqueue               = ti_nqm_rr_enqueue_packet;
    ptr_ti_qos_nqm_mcb->dequeue               = ti_nqm_rr_dequeue_packet;
    ptr_ti_qos_nqm_mcb->congestion_check      = ti_nqm_rr_congestion_check;
    ptr_ti_qos_nqm_mcb->ptr_nqm_private_info  = (void *)ptr_nqm_qos_rr;

    /************* Initialize the Round Robin Scheduler Internals ****************/

    /* Initialize each of the queues in the Network Queue Manager. */
    skb_queue_head_init (&ptr_nqm_qos_rr->nsp_pkt_queue[TI_NSP_QOS_COS1]);
    skb_queue_head_init (&ptr_nqm_qos_rr->nsp_pkt_queue[TI_NSP_QOS_COS2]);
    skb_queue_head_init (&ptr_nqm_qos_rr->nsp_pkt_queue[TI_NSP_QOS_COS3]);
    skb_queue_head_init (&ptr_nqm_qos_rr->nsp_pkt_queue[TI_NSP_QOS_COS4]);
    skb_queue_head_init (&ptr_nqm_qos_rr->nsp_pkt_queue[TI_NSP_QOS_COS5]);
    skb_queue_head_init (&ptr_nqm_qos_rr->nsp_pkt_queue[TI_NSP_QOS_COS6]);

    /* Configure the thresholds of the queues for the various classes. */
    ptr_nqm_qos_rr->nsp_threshold[TI_NSP_QOS_COS1]  = 50;
    ptr_nqm_qos_rr->nsp_threshold[TI_NSP_QOS_COS2]  = 50;
    ptr_nqm_qos_rr->nsp_threshold[TI_NSP_QOS_COS3]  = 50;
    ptr_nqm_qos_rr->nsp_threshold[TI_NSP_QOS_COS4]  = 50;
    ptr_nqm_qos_rr->nsp_threshold[TI_NSP_QOS_COS5]  = 50;
    ptr_nqm_qos_rr->nsp_threshold[TI_NSP_QOS_COS6]  = 50;

    /* Initially there is no congestion and none of the queues are congested. */
    ptr_nqm_qos_rr->nsp_congestion_level = TI_QOS_NQM_NOT_CONGESTED;
    ptr_nqm_qos_rr->nsp_queue_state      = 0x0;

    /* Success: The round-robin scheduler has been initialized. */
    return 0;
}

/**************************************************************************
 * FUNCTION NAME : ti_tca_init
 **************************************************************************
 * DESCRIPTION   :
 *  The function initializes the TCA Configuration. It creates the necessary
 *  proc files that are used by the TCA configuration engine.
 *
 * RETURNS       :
 *          0   -   Success
 *          <0  -   Error
 ***************************************************************************/
static int ti_tca_init (void)
{
    struct proc_dir_entry*	ptr_dir_entry;

    /* Create the PROC Entry used by the TCA Configuration Engine. */
    ptr_dir_entry = create_proc_entry("avalanche/qos" ,0644, NULL);
    if (ptr_dir_entry == NULL)
	{
        printk ("Error: Unable to create QoS proc entry.\n");
    	return -1;
    }
    ptr_dir_entry->data 	  = NULL;
	ptr_dir_entry->read_proc  = ti_tcactl_read_info;
    ptr_dir_entry->write_proc = ti_tcactl_write_info;
	ptr_dir_entry->owner 	  = THIS_MODULE;

    /* Return sucess: Work has been completed. */
    return 0;
}

/**************************************************************************
 * FUNCTION NAME : ti_tca_device_init
 **************************************************************************
 * DESCRIPTION   :
 *  The function is called when a network device is created in the System.
 *  The function allocates memory and initializes the TCA associated with the
 *  device.
 *
 * RETURNS       :
 *  0   -   Success.
 *  <0  -   Error
 ***************************************************************************/
int ti_tca_device_init(struct net_device *dev)
{
    /* Allocate memory for the TCA and initialize it.  */
    dev->ptr_ti_qos_tca = (TI_QOS_TCA *)kmalloc (sizeof(TI_QOS_TCA), GFP_KERNEL);
    if (dev->ptr_ti_qos_tca == NULL)
    {
        printk ("Error: No memory for the TCA.\n");
        return -1;
    }

    /* Initialize the allocated memory block. */
    memset((void *)dev->ptr_ti_qos_tca, 0, sizeof(TI_QOS_TCA));

    /* Reset the Egress TCA to the defaults. */
    ti_tca_reset_egress_tca (dev->ptr_ti_qos_tca);

    /* Reset the Ingress TCA to the defaults. */
    ti_tca_reset_ingress_tca (dev->ptr_ti_qos_tca);

    /* Work has been completed. */
    return 0;
}

/**************************************************************************
 * FUNCTION NAME : ti_tca_device_deinit
 **************************************************************************
 * DESCRIPTION   :
 *  The function is called when a network device is being removed from the 
 *  System.The function cleans the memory associated with the TCA.
 *
 * RETURNS       :
 *  0   -   Success.
 *  <0  -   Error
 ***************************************************************************/
int ti_tca_device_deinit(struct net_device *dev)
{
    /* NSP QoS Framework: If there was a TCA associated with the device. 
     * Relinquish control and clean the memory. */
    if (dev->ptr_ti_qos_tca != NULL)
    {
        /* Clean memory associated with the TCA. */
        kfree (dev->ptr_ti_qos_tca);
    }

    /* Return success. */
    return 0;
}

/**************************************************************************
 * FUNCTION NAME : ti_nsp_high_priority_port_initialization
 **************************************************************************
 * DESCRIPTION   :
 *  The function is called to initialize the HIGH PRiority Port database. 
 *
 * RETURNS       :
 *  0   -   Success.
 *  <0  -   Error
 ***************************************************************************/
static int ti_nsp_high_priority_initialization (void)
{
    int index = 0;

    printk ("DEBUG: Initializing the LOCAL High Priority Database Module. \n");

    /* Initialization of the HIGH Priority Database. */
    ptr_ti_qos_mcb->ptr_local_database = (TI_QOS_LOCAL_DATABASE_MCB *)kmalloc(sizeof(TI_QOS_LOCAL_DATABASE_MCB),GFP_KERNEL);
    if (ptr_ti_qos_mcb->ptr_local_database == NULL)
    {
        printk ("ERROR: Unable to allocate memory for the Local High Priority Database Module.\n");
        return -1;
    }
    
    /* Initialize the port lists. */
    INIT_LIST_HEAD(&ptr_ti_qos_mcb->ptr_local_database->active_port_list);
    INIT_LIST_HEAD(&ptr_ti_qos_mcb->ptr_local_database->available_port_list);

    /* Allocate memory for the available port lists. */
    for (index = 0; index < MAX_NUMBER_LOCAL_PORTS; index++)
    {
        /* Allocate memory for the voice port. */
        TI_QOS_LOCAL_PORT* ptr_voice_port = kmalloc (sizeof(TI_QOS_LOCAL_PORT), GFP_KERNEL);
        if (ptr_voice_port == NULL)
        {
            printk ("ERROR: Unable to allocate memory for the high port nodes.\n");
            return -1;
        }

        /* Initialize the memory block. */
        memset ((void *)ptr_voice_port, 0, sizeof(TI_QOS_LOCAL_PORT));

        /* Add the list of available ports. */
        list_add_tail((struct list_head *)&ptr_voice_port->links, &ptr_ti_qos_mcb->ptr_local_database->available_port_list);
    }

    printk ("DEBUG: Initialization of the LOCAL High Priority Database Module successful..\n");
    return 0;
}

/**************************************************************************
 * FUNCTION NAME : ti_nsp_qos_initialize
 **************************************************************************
 * DESCRIPTION   :
 *  The function initializes the TI QoS NSP Framework.
 ***************************************************************************/
int ti_nsp_qos_initialize (void)
{
    /* Debug Message. */
    printk ("Initializing the TI QoS NSP Framework.\n");

    /* Initialize the TCA Configuration Engine. */
    if (ti_tca_init() < 0)
    {
        printk ("ERROR: Initialization of the TCA Configuration Engine Failed.\n");
        return -1;
    }

    /* Allocate and Initialize memory for the QoS Status reporter. */
    gOutputSection = (char *)kmalloc(MAX_OUTPUT_SECTION_LENGTH, GFP_KERNEL);
    if (gOutputSection == NULL)
    {
        printk ("ERROR: No memory to allocate the QoS Status reporter.\n");
        return -1;
    }
    memset ((void *)gOutputSection, 0, MAX_OUTPUT_SECTION_LENGTH);

    /* Initialize the QOS MCB. */
    ptr_ti_qos_mcb = (TI_QOS_MCB *)kmalloc(sizeof(TI_QOS_MCB), GFP_KERNEL);
    if (ptr_ti_qos_mcb == NULL)
    {
        printk ("ERROR: No memory to allocate the QoS MCB.\n");
        return -1;
    }
    
    /* Initialize the allocated block of memory. */
    memset ((void *)ptr_ti_qos_mcb, 0, sizeof (TI_QOS_MCB));

    /* Initialization of the Network Queue Manager. */
    printk ("DEBUG: Initializing the Network Queue Manager Algorithms.\n");

    /* Allocate memory for the NQM Algorithms. */
    ptr_ti_qos_mcb->ptr_nqm_handle = (TI_QOS_NQM_MCB *)kmalloc(sizeof(TI_QOS_NQM_MCB), GFP_KERNEL);
    if (ptr_ti_qos_mcb->ptr_nqm_handle == NULL)
    {
        printk ("ERROR: No memory to allocate the QoS NQM Algorithms.\n");
        return -1;
    }

    /* Initialize the allocated memory block. */
    memset ((void *)ptr_ti_qos_mcb->ptr_nqm_handle, 0, sizeof (TI_QOS_NQM_MCB));

    /* NSP QoS Framework currently supports the Strict Priority Round-Robin Scheduler. */
    if (ti_nsp_qos_init_round_robin_scheduler(ptr_ti_qos_mcb->ptr_nqm_handle) < 0)
        return -1; 
 
    /* Initialize the Translation Unit. */
    printk ("DEBUG: Initialization of the Translation Unit.\n");

    /* Allocate memory for the Translation Unit. */
    ptr_ti_qos_mcb->ptr_tu_handle  = (TI_QOS_TU_MCB *)kmalloc(sizeof(TI_QOS_TU_MCB), GFP_KERNEL);
    if (ptr_ti_qos_mcb->ptr_tu_handle == NULL)
    {
        printk ("ERROR: Memory allocation of the Translation Unit Failed.\n");
        return -1;
    }

    /* Initialize the Translation Unit Block. */
    memset ((void *)ptr_ti_qos_mcb->ptr_tu_handle, 0, sizeof(TI_QOS_TU_MCB));

    /* Initialize the HIGH Priority Local Database */
    if (ti_nsp_high_priority_initialization() < 0)
        return -1;

    printk ("Succesfully Initialized the TI QoS NSP Framework.\n");
    return 0;
}
