/*--------------------------------------------------------------------------------
 * Copyright 2002-2004, Texas Instruments Incorporated
 *
 * This program has been added from its original operation by Texas Instruments
 * to do the following:
 *       net/sched/sch_priowrr.c : 6-band priority scheduler.
 *
 *   THIS MODIFIED SOFTWARE AND DOCUMENTATION ARE PROVIDED
 *   "AS IS," AND TEXAS INSTRUMENTS MAKES NO REPRESENTATIONS 
 *   OR WARRENTIES, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED 
 *   TO, WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY 
 *   PARTICULAR PURPOSE OR THAT THE USE OF THE SOFTWARE OR 
 *   DOCUMENTATION WILL NOT INFRINGE ANY THIRD PARTY PATENTS, 
 *   COPYRIGHTS, TRADEMARKS OR OTHER RIGHTS. 
 *   See The GNU General Public License for more details.
 *
 * These changes are covered under version 2 of the GNU General Public License, 
 *  dated June 1991.
 *------------------------------------------------------------------------------
 * 
 *  The file contains the implementation of the PRIORITY based WEIGHTED Round
 *  Robin Scheduler. 
 *
 *  It has been updated to handle 6 queues. These queues comprise the Class of
 *  service Queues defined in the NSP QoS Specification.
 *
 *  The 6 queues are as follows:-
 *      a) CoS1
 *      b) CoS2
 *      c) CoS3
 *      d) CoS4
 *      e) CoS5
 *      f) CoS6
 *
 * Enqueuing:-
 *  We have a common pool of buffers for all the queues. The number of buffers 
 *  in the pool corresponds to "tx_queue_len". We need to allocate buffers to 
 *  each queue based on the weight map (or qlen map if configured).
 * i.e. 
 *  SUM (Q(CoS1), Q(CoS2), Q(CoS3), Q(CoS4), Q(CoS5), Q(CoS6))<= tx_queue_len.
 *
 *  Also, the weight of all the classes excluding CoS1 should be 100.
 *  SUM ( W(CoS2), W(CoS3), W(CoS4), W(CoS5), W(CoS6)) = 100
 *
 * Dequeuing:-
 *  The CoS1 queue is serviced to extinction, no preemption. 
 *  All the other queues from CoS2 to CoS6 are serviced (after the CoS1 queue)
 *  based on WRR Queue scheduler.
 * 
 * Traffic Control and priowrr scheduler:-
 *  The "tc" utility is used to enable this scheduler for nas/ppp interfaces, 
 *  configure the scheduler and configure the filters for the scheduler.
 *
 *  Attaching this scheduler to nas/ppp interface:-
 * 	tc qdisc add dev nas0 root handle 1: priowrr [wt1 wt2 wt3 wt4 wt5 wt6] [qlen1 qlen2 qlen3 qlen4 qlen5 qlen6]
 * 		Where wt1 is always 100, sum(wt2,wt3,wt4,wt5,wt6) = 100.
 *		sum(qlen1, qlen2, qlen3, qlen4, qlen5, qlen6) = tx_queue_len.
 *
 * Because of the QoS Framework the Filter rules need NOT be added through the
 * "tc" utility and are NOT supported.
 *************************************************************************/

/**************************** INCLUDE FILES *************************/
#include <linux/config.h>
#include <linux/module.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/bitops.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/socket.h>
#include <linux/sockios.h>
#include <linux/in.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/if_ether.h>
#include <linux/inet.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/notifier.h>
#include <net/ip.h>
#include <net/route.h>
#include <linux/skbuff.h>
#include <net/sock.h>
#include <net/pkt_sched.h>
#include <linux/proc_fs.h>
#include <linux/atmdev.h>

/* NSP QoS Framework. */
#include <linux/ti_qos.h>

typedef struct priowrr_stats_data
{
	int enqueue_stats;
	int dequeue_stats;
	int qlen, maxqlen;
	int drops, flushed;
}PRIOWRR_STATS;

struct priowrr_sched_data
{
	int bands;
	int wted_idx;
	struct tcf_proto *filter_list;	
	u8  weight2band[TCQ_PRIOWRR_BANDS];
	u8  qlen2band[TCQ_PRIOWRR_BANDS];
	struct Qdisc *queues[TCQ_PRIOWRR_BANDS];

    /* Scheduling Table (ST) for WRR scheduler.
     * The Scheduling Table is precomputed for each service slot, that distributes
     * the packets from each of the two Best Effort (BE) queues (Medium and Low) 
     * proportional with the associated values of their weights (wt3, wt4) and 
     * uniformly interleaving them. */
    u8 sch_table[TCQ_MAX_ST];

    PRIOWRR_STATS   stats[TCQ_PRIOWRR_BANDS];
};

/**************************************************************************
 * FUNCTION NAME : priowrr_write_proc
 **************************************************************************
 * DESCRIPTION   :
 *  This is the function that is called after a command is written to the
 *  PRIOWRR proc entry.
 *
 * RETURNS       :
 *  -1              - Error.
 *  Non-Zero        - Success.
 ***************************************************************************/
static int priowrr_write_proc (struct file *file, const char *buffer, unsigned long count, void *data) 
{
	char    priowrr[100];
	char*	argv[10];
	int		argc = 0;
	char*	ptr_cmd;
    char*   delimitters = " \n\t";

	/* Validate the length of data passed. */
	if (count > 100)
		count = 100; 

    /* Initialize the buffer before using it. */
    memset ((void *)&priowrr[0], 0, sizeof(priowrr));
    memset ((void *)&argv[0], 0, sizeof(argv));

	/* Copy from user space. */	
	if (copy_from_user (&priowrr, buffer, count))
		return -EFAULT;

	/* Tokenize the command. Check if there was a NULL entry. If so be the case the
     * user did not know how to use the entry. Print the help screen. */
	ptr_cmd = strtok (priowrr, delimitters);
    if (ptr_cmd == NULL)
        return count;

    /* Parse all the commands typed. */
	do
	{
        /* Extract the first command. */
		argv[argc++] = ptr_cmd;

        /* Validate if the user entered more commands.*/
		if (argc >=10)
		{
			printk ("ERROR: Incorrect too many parameters dropping the command\n");
			return -EFAULT;
		}

        /* Get the next valid command. */
		ptr_cmd = strtok (NULL, delimitters);
	} while (ptr_cmd != NULL);

    /* Show Command Handler. */
    if (strncmp(argv[0], "show", strlen("show")) == 0)
    {
        struct net_device*          dev;
        struct priowrr_sched_data*  ptr_priowrr_data;
        int                         i;

        /* Make sure that the device name is present. */
        if (argv[1] == NULL)
            return -EINVAL;

        /* Validate the device name. */
        dev = dev_get_by_name (argv[1]);
        if (dev == NULL)
            return -EINVAL;

        /* Validate if a queueing discipline was associated or not. */
        if ((dev->qdisc == NULL) || (dev->qdisc->data == NULL) || (dev->qdisc->ops == NULL))
        {
            /* There was no PRIOWRR associated. Release the device and return. */
            printk ("No QDisc associated %s.\n", dev->name);
            dev_put(dev);
            return -EINVAL;
        }

        /* Validate if this is a PRIOWRR Based Scheduling Discipline. */
        if (strcmp(dev->qdisc->ops->id,"priowrr") != 0)
        {
            printk ("No PRIOWRR associated with %s.\n", dev->name);
            dev_put(dev);
            return -EINVAL;
        }

        /* Get the PRIOWRR data. */
        ptr_priowrr_data = (struct priowrr_sched_data *)dev->qdisc->data;

    	printk("Enqueue    Dequeue    Maxqlen    Drops    Flushed\n");
   	    for (i=0; i<TCQ_PRIOWRR_BANDS; i++)
	    {
		    printk("%7d    %7d    %7d    %5d    %7d\n", 
                   ptr_priowrr_data->stats[i].enqueue_stats, ptr_priowrr_data->stats[i].dequeue_stats, 
                   ptr_priowrr_data->stats[i].maxqlen, ptr_priowrr_data->stats[i].drops, 
                   ptr_priowrr_data->stats[i].flushed);
        }

        /* Release the device counter that was held while displaying the
         * statistics. */
        dev_put(dev);

        /* Succesfully processed the command. */
        return count;
    }

    /* Unsupported Command. */
    return -1;
}

/**************************************************************************
 * FUNCTION NAME : priowrr_classify
 **************************************************************************
 * DESCRIPTION   :
 *  The function is the classification routine that determines the queue the
 *  packet is to be placed into depending on the priority of the packet. 
 *  The packet should already have been classified by the QoS Framework hooks
 *  by the time the packet is received in this function. In case the packet
 *  has not been classified the packet is placed into the default CoS6 queue.
 *
 * RETURNS       :
 *  Queue Id associated with one of the 6 TI NSP CoS Queues.
 ***************************************************************************/
static __inline__ unsigned int priowrr_classify(struct sk_buff *skb, struct Qdisc *sch)
{
    TI_QOS_PACKET_INFO *ptr_ti_qos_info;

    /* Get the pointer to the QoS Information field. */
    ptr_ti_qos_info = (TI_QOS_PACKET_INFO *)&skb->qos_pkt_info;

    /* The packet should have passed through the translation unit. If not then this
     * we default the packet to the best effort queue. */
    if (!(ptr_ti_qos_info->ti_qos_flags & TI_QOS_PACKET_TRANSLATED))
    {
        printk ("DEBUG: Packet was NOT passed through the translation unit device:%s skb:0x%p.\n", skb->dev->name, skb);
        ptr_ti_qos_info->ti_service_class = TI_NSP_QOS_COS6;
    }

    /* Return the service class. */
    return (int)ptr_ti_qos_info->ti_service_class;
}

/**************************************************************************
 * FUNCTION NAME : priowrr_enqueue
 **************************************************************************
 * DESCRIPTION   :
 *  The registered enqueue routine with the Scheduler. The function is called
 *  to enqueue a packet into the appropriate queue. The function calls the
 *  'classify' routine to determine the priority of the packet and it then 
 *  places a packet into the appropriate queue.
 *
 * RETURNS       :
 *      0                               - If the function suceeds.
 *      Any other value (NET_XMIT_DROP) - If the queue was FULL.
 ***************************************************************************/
static int priowrr_enqueue(struct sk_buff *skb, struct Qdisc* sch)
{
	struct priowrr_sched_data*  q;
	struct Qdisc*               qdisc;
	int                         ret;
	u32                         queue_id;

    /* Get the private data structure for the PRIOWRR. */
    q = (struct priowrr_sched_data *)sch->data;

    /* Get the Queue Id that is to be used. */
	queue_id = priowrr_classify(skb, sch);

    /* Select the Queueing Discipline associated with the queue id. */
	qdisc = q->queues[queue_id];

	/* Enqueue the packet into the appropriate queueing discipline. */
	if ((ret = qdisc->enqueue(skb, qdisc)) == 0) 	
	{
        /* Succesfully Enqueued the packet. Increment the statistics. */
		sch->stats.bytes += skb->len;
		sch->stats.packets++;
		sch->q.qlen++;
		q->stats[queue_id].enqueue_stats++;
		q->stats[queue_id].qlen++;
		if (q->stats[queue_id].qlen > q->stats[queue_id].maxqlen)
			q->stats[queue_id].maxqlen = q->stats[queue_id].qlen;
		return 0;
	}

    /* The queue was FULL.. Increment the error statistics. */
	sch->stats.drops++;
	q->stats[queue_id].drops++;
	return ret;
}

/**************************************************************************
 * FUNCTION NAME : priowrr_requeue
 **************************************************************************
 * DESCRIPTION   :
 *  The registered requeue routine with the Scheduler.
 *
 * RETURNS       :
 *  Always returns 0.
 ***************************************************************************/
static int priowrr_requeue(struct sk_buff *skb, struct Qdisc* sch)
{
	struct priowrr_sched_data *q = (struct priowrr_sched_data *)sch->data;
	struct Qdisc *qdisc;
	int ret;

	qdisc = q->queues[priowrr_classify(skb, sch)];

	if ((ret = qdisc->ops->requeue(skb, qdisc)) == 0)
    {
		sch->q.qlen++;
		return 0;
	}
	sch->stats.drops++;
	return ret;
}

/**************************************************************************
 * FUNCTION NAME : priowrr_dequeue
 **************************************************************************
 * DESCRIPTION   :
 *  The registered dequeue routine with the Scheduler.The function does the
 *  following:-
 *      a)  Checks if there is a packet in the CoS1 queue. If so the packet
 *          is handled immediately.
 *      b)  In a weighted round robin manner checks all the other queues
 *          from CoS2 to CoS6.
 *
 * RETURNS       :
 *  Pointer to skb  - Packet to be transmitted as determined by the PRIOWRR
 *  NULL            - No Packet to transmit the queues are empty.
 ***************************************************************************/
static struct sk_buff *priowrr_dequeue(struct Qdisc* sch)
{
	struct sk_buff *skb;
	struct priowrr_sched_data *q = (struct priowrr_sched_data *)sch->data;
	int prio, wt_limit;
	struct Qdisc *qdisc;

    /* NSP QoS Framework: Check the CoS1 queue. If there is a packet then
     * process that immediately. Handle the rest of the queues only after
     * this. */
	qdisc = q->queues[TI_NSP_QOS_COS1];
	skb = qdisc->dequeue(qdisc);
	if (skb) 
    {
        /* Yes. The dequeue was succesful. Update the statistics. */
		sch->q.qlen--;
        q->stats[TI_NSP_QOS_COS1].dequeue_stats++;
		q->stats[TI_NSP_QOS_COS1].qlen--;
        return skb;
    }

    /* Run the Weighted Round Robin Algorithm. */
	wt_limit = q->wted_idx;
	do
	{
		prio = q->sch_table[q->wted_idx++];
		if (q->wted_idx>=TCQ_MAX_ST)
			q->wted_idx = 0;
		qdisc = q->queues[prio];
		skb = qdisc->dequeue(qdisc);
		if (skb) 
        {
			sch->q.qlen--;
			q->stats[prio].dequeue_stats++;
			q->stats[prio].qlen--;
			return skb;
		}
	}
	while (q->wted_idx!=wt_limit);

    /* No packet was available in either of the queues. Return NULL. */
	return NULL;
}

/**************************************************************************
 * FUNCTION NAME : priowrr_drop
 **************************************************************************
 * DESCRIPTION   :
 *  The registered drop routine with the Scheduler.
 ***************************************************************************/
static int priowrr_drop(struct Qdisc* sch)
{
	struct priowrr_sched_data *q = (struct priowrr_sched_data *)sch->data;
	int prio;
	struct Qdisc *qdisc;

	for (prio = q->bands-1; prio >= 0; prio--) 
    {
		qdisc = q->queues[prio];
		if (qdisc->ops->drop(qdisc))
        {
			sch->q.qlen--;
			return 1;
		}
	}
	return 0;
}

/**************************************************************************
 * FUNCTION NAME : priowrr_reset
 **************************************************************************
 * DESCRIPTION   :
 *  The registered reset routine with the Scheduler.
 ***************************************************************************/
static void priowrr_reset(struct Qdisc* sch)
{
	int prio;
	struct priowrr_sched_data *q = (struct priowrr_sched_data *)sch->data;

	for (prio=0; prio<q->bands; prio++) 
    {
		q->stats[prio].flushed += q->queues[prio]->q.qlen;
		qdisc_reset(q->queues[prio]);
		q->stats[prio].qlen = q->queues[prio]->q.qlen;
	}
	sch->q.qlen = 0;
}

/**************************************************************************
 * FUNCTION NAME : priowrr_destroy
 **************************************************************************
 * DESCRIPTION   :
 *  The registered destroy routine with the Scheduler.
 ***************************************************************************/
static void priowrr_destroy(struct Qdisc* sch)
{
	int prio;
	struct priowrr_sched_data *q = (struct priowrr_sched_data *)sch->data;

	for (prio=0; prio<q->bands; prio++) 
    {
		qdisc_destroy(q->queues[prio]);
		q->queues[prio] = &noop_qdisc;
	}
	MOD_DEC_USE_COUNT;
	remove_proc_entry("avalanche/priowrr", NULL);
}

/**************************************************************************
 * FUNCTION NAME : new_qdisc_create_dflt
 **************************************************************************
 * DESCRIPTION   :
 *  The function is used to create and initialize the PRIOWRR algorithm and
 *  associating the internal queues with the FIFO Queueing Discipline.
 *
 * RETURNS       :
 *  NULL        - Error.
 *  Non NULL    - Pointer to the allocated queueing discipline.
 ***************************************************************************/
struct Qdisc* new_qdisc_create_dflt(struct net_device *dev, struct Qdisc_ops *ops, struct tc_fifo_qopt *opt, struct Qdisc *qdisc)
{
	struct Qdisc *sch = NULL;
	struct rtattr *rta = NULL;
	int size = sizeof(*sch) + ops->priv_size;
	int len = RTA_LENGTH(sizeof(*opt));

	if (qdisc == NULL)
	{
	    sch = kmalloc(size, GFP_KERNEL);
    	if (!sch)
	    	return NULL;
    	memset(sch, 0, size);

	    skb_queue_head_init(&sch->q);
    	sch->ops = ops;
	    sch->enqueue = ops->enqueue;
    	sch->dequeue = ops->dequeue;
	    sch->dev = dev;
    	sch->stats.lock = &dev->queue_lock;
	    atomic_set(&sch->refcnt, 1);
	}

	/* creating fifo subclasses, with the option for setting qlen limits. */
	rta = kmalloc(len, GFP_KERNEL);
	if (!rta)
		return NULL;
	memset(rta, 0, len);
	rta->rta_type = TCA_OPTIONS;
	rta->rta_len = len;
	memcpy(RTA_DATA(rta), opt, sizeof(*opt));
	
	if (qdisc == NULL)
	{
        if (!ops->init || ops->init(sch, rta) == 0)
		{
		    kfree(rta);
		    return sch;
        }
        kfree(sch);
	}
	else
	{
		if (!ops->change || ops->change(qdisc, rta) == 0)
		{
			kfree(rta);
			return qdisc;
		}
	}
	kfree(rta);
	return NULL;
}

/**************************************************************************
 * FUNCTION NAME : priowrr_tune
 **************************************************************************
 * DESCRIPTION   :
 *  The function is called to initialize and tune the various parameters in
 *  the PRIOWRR.
 *
 * RETURNS       :
 *  0       - Success.
 *  < 0     - Error
 ***************************************************************************/
static int priowrr_tune(struct Qdisc *sch, struct rtattr *opt)
{
	struct priowrr_sched_data *q = (struct priowrr_sched_data *)sch->data;
	struct tc_priowrr_qopt *qopt = RTA_DATA(opt);
	int band, i, total;	
    int index  = 0;
    int number = 0;

    /* Basic Validations: Validate the length of the options. */
	if (opt->rta_len < RTA_LENGTH(sizeof(*qopt)))
		return -EINVAL;

    /* DEBUG Messages: These will be removed. */
    printk ("DEBUG: Weight Maps in the PRIOWRR are\n");
    printk ("CoS1 Weight Map=%d.\n",qopt->weightmap[TI_NSP_QOS_COS1]);
    printk ("CoS2 Weight Map=%d.\n",qopt->weightmap[TI_NSP_QOS_COS2]);
    printk ("CoS3 Weight Map=%d.\n",qopt->weightmap[TI_NSP_QOS_COS3]);
    printk ("CoS4 Weight Map=%d.\n",qopt->weightmap[TI_NSP_QOS_COS4]);
    printk ("CoS5 Weight Map=%d.\n",qopt->weightmap[TI_NSP_QOS_COS5]);
    printk ("CoS6 Weight Map=%d.\n",qopt->weightmap[TI_NSP_QOS_COS6]);

    printk ("DEBUG: Queue Maps in PRIOWRR are\n");
    printk ("CoS1 Queue Length=%d.\n",qopt->qlenmap[TI_NSP_QOS_COS1]);
    printk ("CoS2 Queue Length=%d.\n",qopt->qlenmap[TI_NSP_QOS_COS2]);
    printk ("CoS3 Queue Length=%d.\n",qopt->qlenmap[TI_NSP_QOS_COS3]);
    printk ("CoS4 Queue Length=%d.\n",qopt->qlenmap[TI_NSP_QOS_COS4]);
    printk ("CoS5 Queue Length=%d.\n",qopt->qlenmap[TI_NSP_QOS_COS5]);
    printk ("CoS6 Queue Length=%d.\n",qopt->qlenmap[TI_NSP_QOS_COS6]);

    /* Validations: Make sure that the weightmap passed adds to 100. 
     * Note weightmaps are valid only for the CoS2,CoS3,CoS4,CoS5 and CoS6. */
	total = 0;
	for (i=1; i<TCQ_PRIOWRR_BANDS; i++) 
		total += qopt->weightmap[i];    
	if (total > TCQ_MAX_WT)
		return -EINVAL;

    /* CoS1 has a fixed queue length reserved for it. This is NOT accessible
     * for the other queues. So get the amount of free space that is available 
     * for the rest of the queues. */
	total = sch->dev->tx_queue_len - qopt->qlenmap[TI_NSP_QOS_COS1];

    /* Generate the queue lengths for the rest of the Queues. Provided they
     * have not been able already calculated. */
	if (qopt->qlenmap[TI_NSP_QOS_COS2] == 0)
		qopt->qlenmap[TI_NSP_QOS_COS2] = (total*qopt->weightmap[TI_NSP_QOS_COS2])/100;
	if (qopt->qlenmap[TI_NSP_QOS_COS3] == 0)
		qopt->qlenmap[TI_NSP_QOS_COS3] = (total*qopt->weightmap[TI_NSP_QOS_COS3])/100;
	if (qopt->qlenmap[TI_NSP_QOS_COS4] == 0)
		qopt->qlenmap[TI_NSP_QOS_COS4] = (total*qopt->weightmap[TI_NSP_QOS_COS4])/100;
	if (qopt->qlenmap[TI_NSP_QOS_COS5] == 0)
		qopt->qlenmap[TI_NSP_QOS_COS5] = (total*qopt->weightmap[TI_NSP_QOS_COS5])/100;
	if (qopt->qlenmap[TI_NSP_QOS_COS6] == 0)
		qopt->qlenmap[TI_NSP_QOS_COS6] = (total*qopt->weightmap[TI_NSP_QOS_COS6])/100;

    printk ("DEBUG: After Tuning NSP QoS Framework Queue Length Map, Available Queue Length %d.\n", total);
    printk ("CoS1 Queue Length=%d.\n",qopt->qlenmap[TI_NSP_QOS_COS1]);
    printk ("CoS2 Queue Length=%d.\n",qopt->qlenmap[TI_NSP_QOS_COS2]);
    printk ("CoS3 Queue Length=%d.\n",qopt->qlenmap[TI_NSP_QOS_COS3]);
    printk ("CoS4 Queue Length=%d.\n",qopt->qlenmap[TI_NSP_QOS_COS4]);
    printk ("CoS5 Queue Length=%d.\n",qopt->qlenmap[TI_NSP_QOS_COS5]);
    printk ("CoS6 Queue Length=%d.\n",qopt->qlenmap[TI_NSP_QOS_COS6]);

    /* Create the Scheduling table. */
    memset(q->sch_table, 0, TCQ_MAX_ST);

    /* The scheduling table will have a format as so:- 
     *  [number CoS2][number CoS3][number CoS4][number CoS5][number CoS6]
     *
     * This way CoS2 packets are treated in preferential order at the start of
     * the scheduling table; which follows the NSP QoS Specification. The number of 
     * entries in the scheduling table are 10.*/

    /* Get the number of times CoS2 needs to be scheduled in an iteration
     * of the scheduling table. Place that many number of times in the
     * table. */
    number = (qopt->qlenmap[TI_NSP_QOS_COS2]*TCQ_MAX_ST)/total;
    printk ("DEBUG: Number of times for CoS2: %d.\n", number);
    while (number-- != 0)
        q->sch_table[index++] = TI_NSP_QOS_COS2;

    /* Get the number of times CoS3 needs to be scheduled in an iteration
     * of the scheduling table. Place that many number of times in the
     * table. */
    number = (qopt->qlenmap[TI_NSP_QOS_COS3]*TCQ_MAX_ST)/total;
    printk ("DEBUG: Number of times for CoS3: %d.\n", number);
    while (number-- != 0)
        q->sch_table[index++] = TI_NSP_QOS_COS3;

    /* Get the number of times CoS4 needs to be scheduled in an iteration
     * of the scheduling table. Place that many number of times in the
     * table. */
    number = (qopt->qlenmap[TI_NSP_QOS_COS4]*TCQ_MAX_ST)/total;
    printk ("DEBUG: Number of times for CoS4: %d.\n", number);
    while (number-- != 0)
        q->sch_table[index++] = TI_NSP_QOS_COS4;

    /* Get the number of times CoS5 needs to be scheduled in an iteration
     * of the scheduling table. Place that many number of times in the
     * table. */
    number = (qopt->qlenmap[TI_NSP_QOS_COS5]*TCQ_MAX_ST)/total;
    printk ("DEBUG: Number of times for CoS5: %d.\n", number);
    while (number-- != 0)
        q->sch_table[index++] = TI_NSP_QOS_COS5;

    /* Get the number of times CoS6 needs to be scheduled in an iteration
     * of the scheduling table. Place that many number of times in the
     * table. */
    number = (qopt->qlenmap[TI_NSP_QOS_COS6]*TCQ_MAX_ST)/total;
    printk ("DEBUG: Number of times for CoS6: %d.\n", number);
    while (number-- != 0)
        q->sch_table[index++] = TI_NSP_QOS_COS6;

    /* Validations: Make sure that the Queue Length Maps of all the Queues
     * does NOT exceed the capabilities of the network device. */
	total = 0;
	for (i=0; i<TCQ_PRIOWRR_BANDS; i++)
		total += qopt->qlenmap[i];
	if (total > sch->dev->tx_queue_len)
		return -EINVAL;

    /* Initialize the fields in the PRIOWRR algorithm. */
	sch_tree_lock(sch);
	q->bands = TCQ_PRIOWRR_BANDS;
	q->wted_idx	= 0;
	memcpy(q->weight2band, qopt->weightmap, TCQ_PRIOWRR_BANDS);
	memcpy(q->qlen2band, qopt->qlenmap, TCQ_PRIOWRR_BANDS);
	sch_tree_unlock(sch);

    /* Debug: Dump the current configuration. */
	printk("  weightmap:");
	for (i=0; i<q->bands; i++)
		printk(" %d", q->weight2band[i]);
	printk("\n  qlenmap:");
	for (i=0; i<q->bands; i++)
		printk(" %d", q->qlen2band[i]);
	printk("\n  sch_table:");
	for (i=0; i<TCQ_MAX_ST; i++)
		printk(" %d", q->sch_table[i]);
	printk("\n");

    /* Initialize the PRIOWRR and all the internal queues. The internal queues
     * use the FIFO Queueing Discpline. */
	for (band=0; band<q->bands; band++) 
    {
		struct tc_fifo_qopt pfifo_opt;

        memset(&pfifo_opt, 0, sizeof(pfifo_opt));
		pfifo_opt.limit = q->qlen2band[band];

		if (q->queues[band] == &noop_qdisc) 
        {
			struct Qdisc *child;
			child = new_qdisc_create_dflt(sch->dev, &pfifo_qdisc_ops, &pfifo_opt, NULL);
			if (child) 
            {
				sch_tree_lock(sch);
				child = xchg(&q->queues[band], child);
				if (child != &noop_qdisc)
					qdisc_destroy(child);
				sch_tree_unlock(sch);
			}
		}
		else
		{
			/* Tuning the qlen limits for the queues */
			if (!new_qdisc_create_dflt(sch->dev, &pfifo_qdisc_ops, &pfifo_opt, q->queues[band]))
				return -EINVAL;
		}
	}
	
	for (i=0; i<TCQ_PRIOWRR_BANDS; i++)
	{
		q->stats[i].enqueue_stats  = 0;
		q->stats[i].dequeue_stats  = 0;
		q->stats[i].qlen           = 0;
		q->stats[i].maxqlen        = 0;
		q->stats[i].drops          = 0;
		q->stats[i].flushed        = 0;
	}

    /* Done: Return success. */
	return 0;
}

/**************************************************************************
 * FUNCTION NAME : priowrr_init
 **************************************************************************
 * DESCRIPTION   :
 *  The function is the registered "init" routine.
 *
 * RETURNS       :
 *  0       - Success.
 *  < 0     - Error
 ***************************************************************************/
static int priowrr_init(struct Qdisc *sch, struct rtattr *opt)
{
	struct priowrr_sched_data *q = (struct priowrr_sched_data *)sch->data;
	int i;
    struct proc_dir_entry*	ptr_dir_entry;

	printk("priowrr: %s\n", sch->dev->name);
	for (i=0; i<TCQ_PRIOWRR_BANDS; i++)
		q->queues[i] = &noop_qdisc;

	q->bands = TCQ_PRIOWRR_BANDS;
	memset(q->weight2band, 0, TCQ_PRIOWRR_BANDS);
	memset(q->qlen2band, 0, TCQ_PRIOWRR_BANDS);
	
	if (opt == NULL) {
		return -EINVAL;
	} else {
		int err;

		if ((err= priowrr_tune(sch, opt)) != 0)
			return err;
	}
	MOD_INC_USE_COUNT;

    /* Create the PROC Entry used by the PRIOWRR QoS */
    ptr_dir_entry = create_proc_entry("avalanche/priowrr" ,0644, NULL);
    if (ptr_dir_entry == NULL)
	{
        printk ("Error: Unable to create PRIOWRR proc entry.\n");
        return -1;
    }
    ptr_dir_entry->data 	  = NULL;
    ptr_dir_entry->read_proc  = NULL;
    ptr_dir_entry->write_proc = priowrr_write_proc;
    ptr_dir_entry->owner 	  = THIS_MODULE;

	printk("init: success\n");
	return 0;
}

/**************************************************************************
 * FUNCTION NAME : priowrr_dump
 **************************************************************************
 * DESCRIPTION   :
 *  The function is the registered "dump" routine.
 ***************************************************************************/
static int priowrr_dump(struct Qdisc *sch, struct sk_buff *skb)
{
	struct priowrr_sched_data *q = (struct priowrr_sched_data *)sch->data;
	unsigned char	 *b = skb->tail;
	struct tc_priowrr_qopt opt;

	memcpy(&opt.weightmap, q->weight2band, TCQ_PRIOWRR_BANDS);
	memcpy(&opt.qlenmap, q->qlen2band, TCQ_PRIOWRR_BANDS);
	
	RTA_PUT(skb, TCA_OPTIONS, sizeof(opt), &opt);
	return skb->len;

rtattr_failure:
	skb_trim(skb, b - skb->data);
	return -1;
}

/**************************************************************************
 * FUNCTION NAME : priowrr_graft
 **************************************************************************
 * DESCRIPTION   :
 *  The function is the registered "graft" routine.
 ***************************************************************************/
static int priowrr_graft(struct Qdisc *sch, unsigned long arg, struct Qdisc *new,
		      struct Qdisc **old)
{
	struct priowrr_sched_data *q = (struct priowrr_sched_data *)sch->data;
	unsigned long band = arg - 1;

	if (band >= q->bands)
		return -EINVAL;

	if (new == NULL)
		new = &noop_qdisc;

	sch_tree_lock(sch);
	*old = q->queues[band];
	q->queues[band] = new;
	qdisc_reset(*old);
	sch_tree_unlock(sch);

	return 0;
}

/**************************************************************************
 * FUNCTION NAME : priowrr_leaf
 **************************************************************************
 * DESCRIPTION   :
 *  The function is the registered "leaf" routine.
 ***************************************************************************/
static struct Qdisc *priowrr_leaf(struct Qdisc *sch, unsigned long arg)
{
	struct priowrr_sched_data *q = (struct priowrr_sched_data *)sch->data;
	unsigned long band = arg - 1;

	if (band >= q->bands)
		return NULL;

	return q->queues[band];
}

/**************************************************************************
 * FUNCTION NAME : priowrr_get
 **************************************************************************
 * DESCRIPTION   :
 *  The function is the registered "get" routine.
 ***************************************************************************/
static unsigned long priowrr_get(struct Qdisc *sch, u32 classid)
{
	struct priowrr_sched_data *q = (struct priowrr_sched_data *)sch->data;
	unsigned long band = TC_H_MIN(classid);

	if (band - 1 >= q->bands)
		return 0;
	return band;
}

/**************************************************************************
 * FUNCTION NAME : priowrr_bind
 **************************************************************************
 * DESCRIPTION   :
 *  The function is the registered "bind" routine.
 ***************************************************************************/
static unsigned long priowrr_bind(struct Qdisc *sch, unsigned long parent, u32 classid)
{
	return priowrr_get(sch, classid);
}

/**************************************************************************
 * FUNCTION NAME : priowrr_put
 **************************************************************************
 * DESCRIPTION   :
 *  The function is the registered "put" routine.
 ***************************************************************************/
static void priowrr_put(struct Qdisc *q, unsigned long cl)
{
	return;
}

/**************************************************************************
 * FUNCTION NAME : priowrr_change
 **************************************************************************
 * DESCRIPTION   :
 *  The function is the registered "change" routine.
 ***************************************************************************/
static int priowrr_change(struct Qdisc *sch, u32 handle, u32 parent, struct rtattr **tca, unsigned long *arg)
{
	unsigned long cl = *arg;
	struct priowrr_sched_data *q = (struct priowrr_sched_data *)sch->data;

	if (cl - 1 > q->bands)
		return -ENOENT;
	return 0;
}

/**************************************************************************
 * FUNCTION NAME : priowrr_delete
 **************************************************************************
 * DESCRIPTION   :
 *  The function is the registered "delete" routine.
 ***************************************************************************/
static int priowrr_delete(struct Qdisc *sch, unsigned long cl)
{
	struct priowrr_sched_data *q = (struct priowrr_sched_data *)sch->data;
	if (cl - 1 > q->bands)
		return -ENOENT;
	return 0;
}

/**************************************************************************
 * FUNCTION NAME : priowrr_dump_class
 **************************************************************************
 * DESCRIPTION   :
 *  The function is the registered "dump_class" routine.
 ***************************************************************************/
static int priowrr_dump_class(struct Qdisc *sch, unsigned long cl, struct sk_buff *skb,
			   struct tcmsg *tcm)
{
	struct priowrr_sched_data *q = (struct priowrr_sched_data *)sch->data;

	if (cl - 1 > q->bands)
		return -ENOENT;
	if (q->queues[cl-1])
		tcm->tcm_info = q->queues[cl-1]->handle;
	return 0;
}

/**************************************************************************
 * FUNCTION NAME : priowrr_walk
 **************************************************************************
 * DESCRIPTION   :
 *  The function is the registered "walk" routine.
 ***************************************************************************/
static void priowrr_walk(struct Qdisc *sch, struct qdisc_walker *arg)
{
	struct priowrr_sched_data *q = (struct priowrr_sched_data *)sch->data;
	int prio;

	if (arg->stop)
		return;

	for (prio = 0; prio < q->bands; prio++) {
		if (arg->count < arg->skip) {
			arg->count++;
			continue;
		}
		if (arg->fn(sch, prio+1, arg) < 0) {
			arg->stop = 1;
			break;
		}
		arg->count++;
	}
}

/**************************************************************************
 * FUNCTION NAME : priowrr_find_tcf
 **************************************************************************
 * DESCRIPTION   :
 *  The function is the registered "find TC Filter" routine.
 ***************************************************************************/
static struct tcf_proto ** priowrr_find_tcf(struct Qdisc *sch, unsigned long cl)
{
	struct priowrr_sched_data *q = (struct priowrr_sched_data *)sch->data;

	if (cl)
		return NULL;
	return &q->filter_list;
}

static struct Qdisc_class_ops priowrr_class_ops =
{
	priowrr_graft,
	priowrr_leaf,

	priowrr_get,
	priowrr_put,
	priowrr_change,
	priowrr_delete,
	priowrr_walk,

	priowrr_find_tcf,
	priowrr_bind,
	priowrr_put,

	priowrr_dump_class,
};

struct Qdisc_ops priowrr_qdisc_ops =
{
	NULL,
	&priowrr_class_ops,
	"priowrr",
	sizeof(struct priowrr_sched_data),

	priowrr_enqueue,
	priowrr_dequeue,
	priowrr_requeue,
	priowrr_drop,

	priowrr_init,
	priowrr_reset,
	priowrr_destroy,
	priowrr_tune,

	priowrr_dump,
};

extern struct atm_dev *atm_devs;

/**************************************************************************
 * FUNCTION NAME : flush_driver_queues
 **************************************************************************
 * DESCRIPTION   :
 *  This function is provided to flush the ATM Driver Queues.
 ***************************************************************************/
int flush_driver_queues(int vpi, int vci)
{
	struct tx_flush
	{
		struct atm_vcc *vcc;
		int queue_flag; /* 0, for queue 0, 1 for queue 1, 2 for both */
		int skip_num; 	/* optional parameter to skip "skip_num" of packets from head.*/
	} driver_info;
	struct atm_dev *dev;
	struct atm_vcc *vcc;
		
	memset((void *)&driver_info, 0, sizeof(struct tx_flush));
	
	for (dev = atm_devs; dev; dev = dev->next) {
		for (vcc = dev->vccs; vcc; vcc = vcc->next) {
			if ((vcc->vpi == vpi) && (vcc->vci == vci)) {
				driver_info.vcc = vcc;
				driver_info.queue_flag = 1;
				printk("ATM_FLUSH: vpi/vci %d/%d\n", vpi, vci);
				/* Call ATM driver's ioctl */
				vcc->dev->ops->ioctl (vcc->dev, ATM_FLUSHTX, (void *)&driver_info);
			}
		}
	}
	return 0;
}

#define MIN_MTU		68
/*
 * names[] : device name/s for a PVC. Note that there can be multiple interfaces sharing a single PVC.
 * asymmtu : Asymmetric MTU.
 * vpi,vci : VPI/VCI pair corresponding this PVC, required to call flush ioctl of ATM driver.
 *
 * 	Asymmetric MTU is used for fragmentation to reduce the serialization delay for voice traffic.
 * This is calculated based on the Upstream rate of DSL modem and Interpacket time/delay for 
 * voice packets. This delay is also known as Voice payload size in ms.
 * 
*/
spinlock_t qos_lock;

asmlinkage int sys_config_voice_qos(char *names[], int asymmtu, int vpi, int vci)
{
	int i = 0;
	struct net_device *dev;	
	long flags;
	
	spin_lock_irqsave(&qos_lock, flags);
	
	printk("sys_config_voice:\n");
	
	if ((asymmtu) && (asymmtu < MIN_MTU))
		asymmtu = MIN_MTU;
	
	for (i = 0; names[i] != NULL; i++)
	{
		dev = __dev_get_by_name(names[i]);
		if (!dev)
			continue;
	
		if (asymmtu == 0) {
			/* Reset MTU */
			dev->unused_alignment_fixer = dev->mtu;			
			printk("%s: Reset MTU - %d\n", names[i], dev->unused_alignment_fixer);
		}
		else {			
			/* Set MTU and Flush IP Tx Queues */
			dev->unused_alignment_fixer = ((asymmtu > dev->mtu) ? dev->mtu : asymmtu);
			qdisc_reset(dev->qdisc);			
			printk("%s: Set MTU - %d\n%s: Flush IP queues\n", names[i], dev->unused_alignment_fixer, names[i]);
		}
	}

	if (asymmtu){
		/* Now Flush Driver Tx queues */
		flush_driver_queues(vpi, vci);
	}
	spin_unlock_irqrestore(&qos_lock, flags);
			
	return 0;
}

#ifdef MODULE

int init_module(void)
{
	return register_qdisc(&priowrr_qdisc_ops);
}

void cleanup_module(void) 
{
	unregister_qdisc(&priowrr_qdisc_ops);
}

#endif
MODULE_LICENSE("GPL");

